__all__ = [
    "Loader",
    "Recipe",
    "loader_from",
    "recipe_from",
]
import sys
from dataclasses import dataclass
from enum import auto
from pathlib import Path
from typing import TYPE_CHECKING, Optional, TypedDict, final

import nonos._readers as readers
from nonos._types import BinReader, IniReader, PlanetReader

if sys.version_info >= (3, 11):
    from enum import StrEnum
    from typing import assert_never
else:
    from typing_extensions import assert_never

    from nonos._backports import StrEnum


if TYPE_CHECKING:
    from nonos._types import BinData, IniData, PathT, PlanetData


class Recipe(StrEnum):
    IDEFIX_VTK = auto()
    PLUTO_VTK = auto()
    FARGO3D = auto()
    FARGO_ADSG = auto()


@final
@dataclass(frozen=True, eq=True)
class Loader:
    r"""
    A composable data loader interface.

    Loader instances are immutable and extremely lightweight as they do
    not hold any data other than a Path to a parameter file.
    All actual loading capabilities are deleguated to specialized readers.

    Parameters
    ----------
      parameter_file: Path
        path to an existing parameter file.
      binary_reader: type[BinReader]
        a class that implements the BinReader interface, as defined in nonos._types
      planet_reader: type[PlanetReader]
        a class that implements the PlanetReader interface, as defined in nonos._types
      ini_reader: type[IniReader]
        a class that implements the IniReader interface, as defined in nonos._types

    Raises
    ------
      FileNotFoundError: if `parameter_file` doesn't exist or is a directory.
    """

    # TODO: use slots=True in @dataclass when Python 3.9 is dropped
    __slots__ = [
        "parameter_file",
        "binary_reader",
        "planet_reader",
        "ini_reader",
    ]
    parameter_file: Path
    binary_reader: type["BinReader"]
    planet_reader: type["PlanetReader"]
    ini_reader: type["IniReader"]

    def __post_init__(self) -> None:
        pf = Path(self.parameter_file).resolve()
        if not pf.is_file():
            raise FileNotFoundError(pf)
        object.__setattr__(self, "parameter_file", pf)

    def load_bin_data(self, file: "PathT", /, **meta) -> "BinData":
        ini = self.load_ini_file()
        meta = ini.meta | meta
        return self.binary_reader.read(file, **meta)

    def load_planet_data(self, file: "PathT") -> "PlanetData":
        return self.planet_reader.read(file)

    def load_ini_file(self) -> "IniData":
        return self.ini_reader.read(self.parameter_file)


def loader_from(
    *,
    code: Optional[str] = None,
    parameter_file: Optional["PathT"] = None,
    directory: Optional["PathT"] = None,
) -> Loader:
    r"""
    Compose a Loader object following a known Recipe.

    The exact Recipe needs to be uniquely identifiable from the parameters.

    Parameters
    ----------
      code: str (optional)
        This string should match a Recipe enum member.
        Lower case is expected.
        Valid values include, but are not necessarily limited to:
        - 'idefix_vtk'
        - 'pluto_vtk'
        - 'fargo_adsg'
        - 'fargo3d'

      parameter_file: Path or str (optional)
        A path to a parameter file (e.g. idefix.ini). This path can be
        absolute or relative to the `directory` argument.

      directory: Path or str (optional)
        A path to the simulation directory.

    Raises
    ------
      TypeError: if no argument is provided.
    """
    return _compose_loader(
        recipe_from(
            code=code,
            parameter_file=parameter_file,
            directory=directory,
        ),
        _parameter_file_from(
            parameter_file=parameter_file,
            directory=directory,
        ),
    )


def _compose_loader(recipe: Recipe, /, parameter_file: Path) -> Loader:
    return Loader(parameter_file, **_ingredients_from(recipe))


def _parameter_file_from(
    *,
    parameter_file: Optional["PathT"] = None,
    directory: Optional["PathT"] = None,
) -> Path:
    if parameter_file is None and directory is None:
        raise TypeError(
            "Missing required keyword arguments: 'parameter_file', 'directory' "
            "(need at least one)"
        )

    if directory is not None:
        directory = Path(directory).resolve()
        if parameter_file is None:
            return _parameter_file_from_dir(directory)

    if parameter_file is not None:
        parameter_file = Path(parameter_file)
        if parameter_file.is_absolute():
            return parameter_file
        elif directory is not None and parameter_file == Path(parameter_file.name):
            return directory / parameter_file

    raise ValueError(
        f"Received apparently inconsistent inputs {parameter_file=} and {directory=}"
    )


def _parameter_file_from_dir(directory: "PathT", /) -> Path:
    directory = Path(directory).resolve()
    candidates = list(directory.glob("*.ini"))
    candidates.extend(list(directory.glob("*.par")))
    if len(candidates) == 1:
        return candidates[0]
    elif not candidates:
        raise FileNotFoundError(f"Could not find a parameter file in {directory}")
    else:
        raise RuntimeError(
            f"Found multiple parameter files in {directory}\n - "
            + "\n - ".join(str(c) for c in candidates)
        )


class Ingredients(TypedDict):
    binary_reader: type[BinReader]
    planet_reader: type[PlanetReader]
    ini_reader: type[IniReader]


def _ingredients_from(recipe: Recipe, /) -> Ingredients:
    if recipe is Recipe.IDEFIX_VTK:
        return {
            "binary_reader": readers.binary.VTKReader,
            "planet_reader": readers.planet.IdefixReader,
            "ini_reader": readers.ini.IdefixVTKReader,
        }
    elif recipe is Recipe.PLUTO_VTK:
        return {
            "binary_reader": readers.binary.VTKReader,
            "planet_reader": readers.planet.NullReader,
            "ini_reader": readers.ini.PlutoVTKReader,
        }
    elif recipe is Recipe.FARGO3D:
        return {
            "binary_reader": readers.binary.Fargo3DReader,
            "planet_reader": readers.planet.Fargo3DReader,
            "ini_reader": readers.ini.Fargo3DReader,
        }
    elif recipe is Recipe.FARGO_ADSG:
        return {
            "binary_reader": readers.binary.FargoADSGReader,
            "planet_reader": readers.planet.FargoADSGReader,
            "ini_reader": readers.ini.FargoADSGReader,
        }
    else:
        assert_never(recipe)


def recipe_from(
    *,
    code: Optional[str] = None,
    parameter_file: Optional["PathT"] = None,
    directory: Optional["PathT"] = None,
) -> Recipe:
    r"""
    Determine an appropriate loader recipe from user input.

    Parameters
    ----------
      code: str (optional)
        This string should match a Recipe enum member.
        Lower case is expected.
        Valid values include, but are not necessarily limited to:
        - 'idefix_vtk'
        - 'pluto_vtk'
        - 'fargo_adsg'
        - 'fargo3d'

      parameter_file: Path or str (optional)
        A path to a parameter file (e.g. idefix.ini). This path can be
        absolute or relative to the `directory` argument.

      directory: Path or str (optional)
        A path to the simulation directory.

    Returns
    -------
       a Recipe enum member

    Raises
    ------
      TypeError: if no argument is provided.

      ValueError: if `code` is omitted and a working inifile reader cannot
        be uniquely identified.
    """
    if code is not None:
        return _code_to_recipe(code)

    parameter_file = _parameter_file_from(
        parameter_file=parameter_file,
        directory=directory,
    )

    recipe_candidates: list[Recipe] = []
    for recipe in Recipe.__members__.values():
        loader = _compose_loader(recipe, parameter_file)
        try:
            loader.load_ini_file()
        except Exception:
            continue
        else:
            recipe_candidates.append(recipe)
    if len(recipe_candidates) == 1:
        return recipe_candidates[0]
    elif len(recipe_candidates) == 0:
        msg = (
            f"Could not determine data format from {parameter_file=!r} "
            "(failed to read with any loader)"
        )
    else:  # pragma: no cover
        msg = (
            f"Could not determine unambiguous data format from {parameter_file=!r} "
            f"(found {len(recipe_candidates)} candidates {recipe_candidates})"
        )

    raise ValueError(msg)


def _code_to_recipe(code: str, /) -> Recipe:
    if code in ("pluto", "idefix"):
        # backward compatibility layer
        # this could be deprecated at some point
        code += "_vtk"
    return Recipe(code)



import pathlib
import sys
from io import StringIO

import pytest
from cogapp import Cog


# TODO: remove this marker when Python 3.13 is finalized
@pytest.mark.skipif(
    sys.version_info >= (3, 13),
    reason="argparse --help messages were slightly improved in Python 3.13",
)
@pytest.mark.skipif(
    sys.platform.startswith("win"),
    reason="Windows runner may choke on some special chararcters",
)
@pytest.mark.skipif(
    sys.version_info < (3, 10),
    reason="argparse --help messages were slightly modified in Python 3.10",
)
def test_if_cog_needs_to_be_run():
    _stdout = sys.stdout
    sys.stdout = StringIO()
    readme = pathlib.Path(__file__).parents[1] / "README.md"

    Cog().main(["cog", str(readme)])

    output = sys.stdout.getvalue()
    sys.stdout = _stdout
    assert (
        output == readme.read_text()
    ), "Run 'cog -r README.md' from the top level of the repo (with Python >= 3.10)"



import warnings
from collections.abc import Callable
from math import isclose
from typing import TYPE_CHECKING, Optional

if TYPE_CHECKING:
    # note that these are documented as deprecated, but the replacement
    # (typing.Protocol) isn't flexible enough here (I think)
    # see https://mypy.readthedocs.io/en/stable/additional_features.html#extended-callable-types
    from mypy_extensions import DefaultArg, DefaultNamedArg

    find_phip_T = Callable[
        [DefaultArg(Optional[int]), DefaultNamedArg(Optional[str], "planet_file")],
        float,
    ]


def _parse_planet_file(
    *,
    planet_file: Optional[str] = None,
    planet_number: Optional[int] = None,
) -> str:
    if planet_number is not None and planet_file is not None:
        raise TypeError(
            "received both planet_number and planet_file arguments. "
            "Please pass at most one."
        )
    if planet_file is not None:
        return planet_file
    else:
        return f"planet{planet_number or 0}.dat"


def _parse_rotation_angle(
    *,
    rotate_by: Optional[float],
    rotate_with: Optional[str],
    planet_number_argument: tuple[str, Optional[int]],
    find_phip: "find_phip_T",
    stacklevel: int,
) -> float:
    planet_number_argname, planet_number = planet_number_argument
    defined_args = {rotate_with, rotate_by, planet_number} - {None}
    if not defined_args:
        # no rotation specified
        return 0.0

    if len(defined_args) > 1:
        raise TypeError(
            "Can only process one argument out of "
            f"(rotate_by, rotate_with, {planet_number_argname})"
        )

    # beyond this point, we know that exactly one parameter was specified,
    # let's funnel it down to a rotate_by form
    if planet_number is not None:
        warnings.warn(
            f"The {planet_number_argname} argument is deprecated and will be removed "
            "in a future version. Instead, please use either rotate_by (float) "
            "or rotate_with (str path to planet log file).",
            DeprecationWarning,
            stacklevel=stacklevel + 1,
        )
        rotate_with = _parse_planet_file(planet_number=planet_number)

    if rotate_with is not None:
        rotate_by = find_phip(planet_file=rotate_with)

    if rotate_by is None:  # pragma: no cover
        # this is never supposed to happen, but it's needed to convince mypy that
        # we will not return a None
        raise RuntimeError("Something went terribly wrong. Please report this.")

    return rotate_by


def _fequal(a: float, b: float, /) -> bool:
    # a fuzzy single-precision floating point comparison
    return isclose(a, b, abs_tol=1e-7, rel_tol=1e-6)



import dataclasses
import json
import warnings
from collections import deque
from collections.abc import ItemsView, KeysView, ValuesView
from functools import cached_property
from pathlib import Path
from shutil import copyfile
from typing import TYPE_CHECKING, Any, Optional, Union, overload

import numpy as np
from matplotlib.scale import SymmetricalLogTransform
from matplotlib.ticker import SymmetricalLogLocator

from nonos._readers.binary import NPYReader
from nonos._types import PathT, PlanetData
from nonos.api._angle_parsing import (
    _fequal,
    _parse_planet_file,
    _parse_rotation_angle,
)
from nonos.api.tools import find_around, find_nearest
from nonos.loaders import Recipe, loader_from, recipe_from
from nonos.logging import logger

if TYPE_CHECKING:
    from matplotlib.axes import Axes
    from matplotlib.figure import Figure


class Plotable:
    def __init__(self, dict_plotable: dict) -> None:
        self.dict_plotable = dict_plotable
        self.data = self.dict_plotable[self.dict_plotable["field"]]
        self.dimension = len(self.data.shape)
        if self.dimension > 2:
            raise TypeError(
                "Plotable doesn't support data with dimensionality>2, "
                f"got {self.dimension}"
            )

    def plot(
        self,
        fig: "Figure",
        ax: "Axes",
        *,
        log=False,
        cmap="inferno",
        filename=None,
        fmt="png",
        dpi=500,
        title=None,
        unit_conversion=None,
        nbin=None,  # deprecated
        **kwargs,
    ):
        if nbin is not None:
            warnings.warn(
                "The nbin parameter has no effect and is deprecated",
                stacklevel=2,
            )
        data = self.data
        if unit_conversion is not None:
            data = data * unit_conversion
        if log:
            data = np.log10(data)

        if self.dimension == 2:
            self.akey = self.dict_plotable["abscissa"]
            self.okey = self.dict_plotable["ordinate"]
            self.avalue = self.dict_plotable[self.akey]
            self.ovalue = self.dict_plotable[self.okey]
            kw = {}
            if (norm := kwargs.get("norm")) is not None:
                if "vmin" in kwargs:
                    norm.vmin = kwargs.pop("vmin")
                if "vmax" in kwargs:
                    norm.vmax = kwargs.pop("vmax")
            else:
                vmin = kwargs.pop("vmin") if "vmin" in kwargs else np.nanmin(data)
                vmax = kwargs.pop("vmax") if "vmax" in kwargs else np.nanmax(data)
                kw.update({"vmin": vmin, "vmax": vmax})

            im = ax.pcolormesh(
                self.avalue,
                self.ovalue,
                data,
                cmap=cmap,
                **kwargs,
                **kw,
            )
            ax.set(
                xlim=(self.avalue.min(), self.avalue.max()),
                ylim=(self.ovalue.min(), self.ovalue.max()),
            )

            ax.set_xlabel(self.akey)
            ax.set_ylabel(self.okey)
            if title is not None:
                from mpl_toolkits.axes_grid1 import make_axes_locatable

                divider = make_axes_locatable(ax)
                cax = divider.append_axes("right", size="5%", pad=0.05)
                cbar = fig.colorbar(
                    im, cax=cax, orientation="vertical"
                )  # , format='%.0e')
                cbar.set_label(title)

                cb_axis = cbar.ax.yaxis
                if cb_axis.get_scale() == "symlog":
                    # no minor tick is drawn in symlog norms by default
                    # as of matplotlib 3.7.1, see
                    # https://github.com/matplotlib/matplotlib/issues/25994
                    trf = cb_axis.get_transform()
                    if not isinstance(trf, SymmetricalLogTransform):
                        raise RuntimeError
                    cb_axis.set_major_locator(SymmetricalLogLocator(trf))
                    if float(trf.base).is_integer():
                        locator = SymmetricalLogLocator(
                            trf, subs=list(range(1, int(trf.base)))
                        )
                        cb_axis.set_minor_locator(locator)
            else:
                return im
        elif self.dimension == 1:
            vmin = kwargs.pop("vmin") if "vmin" in kwargs else np.nanmin(data)
            vmax = kwargs.pop("vmax") if "vmax" in kwargs else np.nanmax(data)
            self.akey = self.dict_plotable["abscissa"]
            self.avalue = self.dict_plotable[self.akey]
            if "norm" in kwargs:
                logger.info("norm has no meaning in 1D.")
                kwargs.pop("norm")
            ax.plot(self.avalue, data, **kwargs)
            ax.set_ylim(ymin=vmin)
            ax.set_ylim(ymax=vmax)
            ax.set_xlabel(self.akey)
            if title is not None:
                ax.set_ylabel(title)
        else:
            raise TypeError(
                "Plotable doesn't support data with dimensionality>2, "
                f"got {self.dimension}"
            )
        if filename is not None:
            fig.savefig(f"{filename}.{fmt}", bbox_inches="tight", dpi=dpi)


class Coordinates:
    """Coordinates class from x1, x2, x3"""

    def __init__(
        self, geometry: str, x1: np.ndarray, x2: np.ndarray, x3: np.ndarray
    ) -> None:
        if x1.shape[0] == 1:
            x1 = np.array([x1[0], x1[0]])
        if x2.shape[0] == 1:
            x2 = np.array([x2[0], x2[0]])
        if x3.shape[0] == 1:
            x3 = np.array([x3[0], x3[0]])
        self.geometry = geometry
        if self.geometry == "cartesian":
            self.x = x1
            self.y = x2
            self.z = x3
            self.cube = ("x", "y", "z")
            self.xmed = 0.5 * (self.x[1:] + self.x[:-1])
            self.ymed = 0.5 * (self.y[1:] + self.y[:-1])
            self.zmed = 0.5 * (self.z[1:] + self.z[:-1])
        if self.geometry == "spherical":
            self.r = x1
            self.theta = x2
            self.phi = x3
            self.cube = ("r", "theta", "phi")
            self.rmed = 0.5 * (self.r[1:] + self.r[:-1])
            self.thetamed = 0.5 * (self.theta[1:] + self.theta[:-1])
            self.phimed = 0.5 * (self.phi[1:] + self.phi[:-1])
        if self.geometry == "polar":
            self.R = x1
            self.phi = x2
            self.z = x3
            self.cube = ("R", "phi", "z")
            self.Rmed = 0.5 * (self.R[1:] + self.R[:-1])
            self.phimed = 0.5 * (self.phi[1:] + self.phi[:-1])
            self.zmed = 0.5 * (self.z[1:] + self.z[:-1])

    @property
    def shape(self) -> tuple[int, int, int]:
        """
        Returns
        =======
        shape : tuple
        """
        if self.geometry == "cartesian":
            return len(self.x), len(self.y), len(self.z)
        elif self.geometry == "spherical":
            return len(self.r), len(self.theta), len(self.phi)
        elif self.geometry == "polar":
            return len(self.R), len(self.phi), len(self.z)
        else:
            raise RuntimeError(f"Unknown geometry {self.geometry!r}")

    @property
    def get_attributes(self) -> dict[str, Any]:
        if self.geometry == "cartesian":
            return {"geometry": self.geometry, "x": self.x, "y": self.y, "z": self.z}
        elif self.geometry == "spherical":
            return {
                "geometry": self.geometry,
                "r": self.r,
                "theta": self.theta,
                "phi": self.phi,
            }
        elif self.geometry == "polar":
            return {
                "geometry": self.geometry,
                "R": self.R,
                "phi": self.phi,
                "z": self.z,
            }
        else:
            raise RuntimeError(f"Unknown geometry {self.geometry!r}")

    @property
    def get_coords(self) -> dict[str, Any]:
        if self.geometry == "cartesian":
            return {
                "x": self.x,
                "y": self.y,
                "z": self.z,
                "xmed": self.xmed,
                "ymed": self.ymed,
                "zmed": self.zmed,
            }
        elif self.geometry == "spherical":
            return {
                "r": self.r,
                "theta": self.theta,
                "phi": self.phi,
                "rmed": self.rmed,
                "thetamed": self.thetamed,
                "phimed": self.phimed,
            }
        elif self.geometry == "polar":
            return {
                "R": self.R,
                "phi": self.phi,
                "z": self.z,
                "Rmed": self.Rmed,
                "phimed": self.phimed,
                "zmed": self.zmed,
            }
        else:
            raise RuntimeError(f"Unknown geometry {self.geometry!r}")

    def _meshgrid_reduction(self, *reducted) -> dict:
        for i in reducted:
            if i not in self.cube:
                raise KeyError(f"{i} not in {self.cube}")
        dictcoords = {}
        if len(reducted) <= 2:
            for coords in reducted:
                dictcoords[coords] = vars(self)[coords]
            axis = list(set(reducted) ^ set(self.cube))
            dictmesh: dict[str, Any] = {}
            # 2D map
            if len(axis) == 1:
                dictmesh[reducted[0]], dictmesh[reducted[1]] = np.meshgrid(
                    dictcoords[reducted[0]], dictcoords[reducted[1]]
                )
                axismed = "".join([axis[0], "med"])
                dictmesh[axis[0]] = vars(self)[axismed]
                # carefule: takes "xy", "yz", "zx" (all combinations)
                if "".join(reducted) in "".join((*self.cube, self.cube[0])):
                    ordered = True
                else:
                    ordered = False
                dictmesh["ordered"] = ordered
            # 1D curve
            else:
                dictmesh[reducted[0]] = vars(self)["".join([reducted[0], "med"])]
        else:
            raise ValueError(f"more than 2 coordinates were specified: {reducted}.")
        return dictmesh

    # on demande 'x','y' et la geometry est cartesian -> 'x','y'
    # on demande 'x','y' et la geometry est polaire -> 'R','phi'
    # on demande 'x','y' et la geometry est spherique -> 'r','phi'
    @overload
    def native_from_wanted(
        self, _wanted_x1: str, _wanted_x2: str, /
    ) -> tuple[tuple[str, str], str]: ...

    @overload
    def native_from_wanted(
        self, _wanted_x1: str, _wanted_x2: None, /
    ) -> tuple[tuple[str], str]: ...

    def native_from_wanted(self, _wanted_x1: str, _wanted_x2: Optional[str] = None, /):
        if self.geometry == "cartesian":
            conversion = {
                "x": "x",
                "y": "y",
                "z": "z",
            }
        elif self.geometry == "polar":
            conversion = {
                "R": "R",
                "phi": "phi",
                "z": "z",
                "x": "R",
                "y": "phi",
                "r": "R",
                "theta": "z",
            }
        elif self.geometry == "spherical":
            conversion = {
                "r": "r",
                "theta": "theta",
                "phi": "phi",
                "x": "r",
                "y": "phi",
                "z": "theta",
                "R": "r",
            }
        else:
            raise RuntimeError(f"Unknown geometry {self.geometry!r}")

        wanted: tuple[str, ...]
        if _wanted_x2 is None:
            wanted = (_wanted_x1,)
        else:
            wanted = (_wanted_x1, _wanted_x2)
        for i in wanted:
            if i not in conversion:
                raise KeyError(f"{i} not in {tuple(conversion.keys())}")
        if {"x", "y", "z"}.issuperset(wanted):
            target_geometry = "cartesian"
        elif {"R", "phi", "z"}.issuperset(wanted):
            target_geometry = "polar"
        elif {"r", "theta", "phi"}.issuperset(wanted):
            target_geometry = "spherical"
        else:
            raise ValueError(f"Unknown wanted plane: {wanted}.")

        native: tuple[str, ...]
        if _wanted_x2 is None:
            native = (conversion[_wanted_x1],)
        else:
            native = (conversion[_wanted_x1], conversion[_wanted_x2])

        return native, target_geometry

    # for 2D arrays
    def target_from_native(self, target_geometry, coords) -> dict[str, np.ndarray]:
        if self.geometry == "polar":
            R, phi, z = (coords["R"], coords["phi"], coords["z"])
            if target_geometry == "cartesian":
                x = R * np.cos(phi)
                y = R * np.sin(phi)
                target_coords = {"x": x, "y": y, "z": z}
            elif target_geometry == "spherical":
                r = np.sqrt(R**2 + z**2)
                theta = np.arctan2(R, z)
                target_coords = {"r": r, "theta": theta, "phi": phi}
            else:
                raise ValueError(f"Unknown target geometry {target_geometry}.")

        elif self.geometry == "cartesian":
            x, y, z = (coords["x"], coords["y"], coords["z"])
            if target_geometry == "polar":
                R = np.sqrt(x**2 + y**2)
                phi = np.arctan2(y, x)
                target_coords = {"R": R, "phi": phi, "z": z}
            elif target_geometry == "spherical":
                r = np.sqrt(x**2 + y**2 + z**2)
                theta = np.arctan2(np.sqrt(x**2 + y**2), z)
                phi = np.arctan2(y, x)
                target_coords = {"r": r, "theta": theta, "phi": phi}
            else:
                raise ValueError(f"Unknown target geometry {target_geometry}.")

        elif self.geometry == "spherical":
            r, theta, phi = (coords["r"], coords["theta"], coords["phi"])
            if target_geometry == "polar":
                R = r * np.sin(theta)
                z = r * np.cos(theta)
                target_coords = {"R": R, "phi": phi, "z": z}
            elif target_geometry == "cartesian":
                x = r * np.sin(theta) * np.cos(phi)
                y = r * np.sin(theta) * np.sin(phi)
                z = r * np.cos(theta)
                if len(theta.shape) <= 1:
                    x = r * np.sin(theta) * np.cos(phi)
                    y = r * np.sin(theta) * np.sin(phi)
                    z = np.cos(theta)
                target_coords = {"x": x, "y": y, "z": z}
            else:
                raise ValueError(f"Unknown target geometry {target_geometry}.")

        else:
            raise ValueError(f"Unknown geometry {self.geometry}.")

        target_coords["ordered"] = coords["ordered"]
        return target_coords

    def _meshgrid_conversion(self, *wanted) -> dict:
        native_from_wanted = self.native_from_wanted(*wanted)
        native = native_from_wanted[0]
        target_geometry = native_from_wanted[1]
        native_meshcoords = self._meshgrid_reduction(*native)
        if len(wanted) == 1:
            return native_meshcoords
        else:
            meshcoords = {}
            if target_geometry == self.geometry:
                meshcoords = native_meshcoords
            else:
                meshcoords = self.target_from_native(target_geometry, native_meshcoords)
            return meshcoords


class GasField:
    """Idefix field class

    Attributes
    ==========
    data : array corresponding to the 3D data cube.
    coords : Coordinates of the data.
        Edge of the cells.
    coordsmed : Coordinates of the data.
        Center of the cells.
    """

    def __init__(
        self,
        field: str,
        data: np.ndarray,
        coords: Coordinates,
        ngeom: str,
        on: int,
        operation: str,
        *,
        inifile: Optional[PathT] = None,
        code: Union[str, Recipe, None] = None,
        directory: Optional[PathT] = None,
        rotate_by: Optional[float] = None,
        rotate_with: Optional[str] = None,
        rotate_grid: int = -1,  # deprecated
    ) -> None:
        self.field = field
        self.operation = operation
        self.native_geometry = ngeom
        self.data = data
        self.coords = coords
        self.on = on

        if directory is None:
            directory = Path.cwd()
        self.directory = Path(directory)

        self._loader = loader_from(
            code=code,
            parameter_file=inifile,
            directory=directory,
        )
        self.inifile = self._loader.parameter_file

        self._rotate_by = _parse_rotation_angle(
            rotate_by=rotate_by,
            rotate_with=rotate_with,
            planet_number_argument=(
                "rotate_grid",
                (rotate_grid if rotate_grid >= 0 else None),
            ),
            stacklevel=2,
            find_phip=self.find_phip,
        )

        # TODO: remove this after deprecation
        self._recipe = recipe_from(parameter_file=inifile, directory=directory)
        self._code = str(code or self._recipe)

    @property
    def code(self) -> str:
        warnings.warn(
            "GasField.code is deprecated and will be removed in a future version.",
            category=DeprecationWarning,
            stacklevel=2,
        )
        return self._code

    @property
    def shape(self) -> tuple[int, int, int]:
        """
        Returns
        =======
        shape : tuple
        """
        i, j, k = (max(1, n - 1) for n in self.coords.shape)
        return i, j, k

    def map(
        self,
        *wanted,
        rotate_by: Optional[float] = None,
        rotate_with: Optional[str] = None,
        planet_corotation: Optional[int] = None,  # deprecated
    ) -> Plotable:
        rotate_by = _parse_rotation_angle(
            rotate_by=rotate_by,
            rotate_with=rotate_with,
            planet_number_argument=("planet_corotation", planet_corotation),
            stacklevel=2,
            find_phip=self.find_phip,
        )

        data_key = self.field
        # we count the number of 1 in the shape of the data, which gives the real dimension of the data,
        # i.e. the number of reductions already performed (0 -> 3D, 1 -> 2D, 2 -> 1D)
        if self.shape.count(1) not in (1, 2):
            raise ValueError("data has to be 1D or 2D in order to call map.")
        dimension = len(wanted)

        meshgrid_conversion = self.coords._meshgrid_conversion(*wanted)

        if dimension == 1:
            abscissa_value = list(meshgrid_conversion.values())[0]
            abscissa_key = list(meshgrid_conversion.keys())[0]
            if "phi" in wanted and not _fequal(self._rotate_by, rotate_by):
                phicoord = self.coords.phi - rotate_by
                ipi = find_nearest(phicoord, 0)
                if abs(0 - phicoord[ipi]) > abs(
                    np.ediff1d(find_around(phicoord, 0))[0]
                ):
                    ipi = find_nearest(phicoord, 2 * np.pi)
                if self.native_geometry == "polar":
                    data_view = np.roll(self.data, -ipi + 1, axis=1)
                elif self.native_geometry == "spherical":
                    data_view = np.roll(self.data, -ipi + 1, axis=2)
                else:
                    raise NotImplementedError(
                        f"geometry flag '{self.native_geometry}' not implemented yet if corotation"
                    )
            else:
                data_view = self.data.view()

            dict_plotable = {
                "abscissa": abscissa_key,
                "field": data_key,
                abscissa_key: abscissa_value,
                data_key: data_view.squeeze(),
            }
        elif dimension == 2:
            # meshgrid in polar coordinates P, R (if "R", "phi") or R, P (if "phi", "R")
            # idem for all combinations of R,phi,z
            abscissa_value, ordinate_value = (
                meshgrid_conversion[wanted[0]],
                meshgrid_conversion[wanted[1]],
            )
            abscissa_key, ordinate_key = (wanted[0], wanted[1])
            native_from_wanted = self.coords.native_from_wanted(*wanted)[0]
            if "phi" in native_from_wanted and not _fequal(self._rotate_by, rotate_by):
                phicoord = self.coords.phi - rotate_by
                # ipi = find_nearest(phicoord, np.pi)
                # if (abs(np.pi-phicoord[ipi])>abs(np.ediff1d(find_around(phicoord, np.pi))[0])):
                #     ipi = find_nearest(phicoord, -np.pi)
                ipi = find_nearest(phicoord, 0)
                if abs(0 - phicoord[ipi]) > abs(
                    np.ediff1d(find_around(phicoord, 0))[0]
                ):
                    ipi = find_nearest(phicoord, 2 * np.pi)
                if self.native_geometry == "polar":
                    data_view = np.roll(self.data, -ipi + 1, axis=1)
                elif self.native_geometry == "spherical":
                    data_view = np.roll(self.data, -ipi + 1, axis=2)
                else:
                    raise NotImplementedError(
                        f"geometry flag '{self.native_geometry}' not implemented yet if corotation"
                    )
            else:
                data_view = self.data.view()

            def rotate_axes(arr, shift: int):
                axes_in = tuple(range(arr.ndim))
                axes_out = deque(axes_in)
                axes_out.rotate(shift)
                return np.moveaxis(arr, axes_in, axes_out)

            # make reduction axis the first axis then drop (squeeze) it,
            # while preserving the original (cyclic) order in the other two axes
            data_view = rotate_axes(data_view, shift=self.shape.index(1)).squeeze()

            if meshgrid_conversion["ordered"]:
                data_view = data_view.T

            dict_plotable = {
                "abscissa": abscissa_key,
                "ordinate": ordinate_key,
                "field": data_key,
                abscissa_key: abscissa_value,
                ordinate_key: ordinate_value,
                data_key: data_view,
            }
        else:
            raise RuntimeError

        assert dict_plotable[data_key].ndim == dimension

        return Plotable(dict_plotable)

    def save(
        self,
        directory: Optional[PathT] = None,
        header_only: bool = False,
    ) -> Path:
        if directory is None:
            directory = Path.cwd()
        else:
            directory = Path(directory)
        operation = self.operation
        headerdir = directory / "header"
        subdir = directory / self.field.lower()
        file = subdir / f"{operation}_{self.field}.{self.on:04d}.npy"
        if not header_only:
            if file.is_file():
                logger.info("{} already exists", file)
            else:
                subdir.mkdir(exist_ok=True, parents=True)
                with open(file, "wb") as fh:
                    np.save(fh, self.data)

        group_of_files = list(subdir.glob(f"{operation}*"))
        op_suffix = f"_{operation}" if operation != "" else ""
        filename = f"header{op_suffix}.json"
        header_file = headerdir / filename
        if (len(group_of_files) > 0 and not header_file.is_file()) or header_only:
            headerdir.mkdir(exist_ok=True, parents=True)
            if header_file.is_file():
                logger.info("{} already exists", header_file)
            else:
                dictsaved = self.coords.get_attributes

                def is_array(item: tuple[str, Any]) -> bool:
                    _key, value = item
                    return isinstance(value, np.ndarray)

                for key, value in filter(is_array, dictsaved.items()):
                    dictsaved[key] = value.tolist()
                with open(header_file, "w") as hfile:
                    json.dump(dictsaved, hfile, indent=2)

        src = self.inifile.resolve()
        dest = directory / self.inifile.name
        if dest != src:
            copyfile(src, dest)

        return file

    def find_ir(self, distance=1.0):
        if self.native_geometry in ("polar"):
            return find_nearest(self.coords.Rmed, distance)
        if self.native_geometry in ("spherical"):
            return find_nearest(self.coords.rmed, distance)

    def find_imid(self, altitude=0.0):
        if self.native_geometry in ("cartesian", "polar"):
            return find_nearest(self.coords.zmed, altitude)
        if self.native_geometry in ("spherical"):
            return find_nearest(self.coords.thetamed, np.pi / 2 - altitude)

    def find_iphi(self, phi=0):
        if self.native_geometry in ("polar", "spherical"):
            mod = len(self.coords.phi) - 1
            return find_nearest(self.coords.phi, phi) % mod

    def _load_planet(
        self,
        *,
        planet_number: Optional[int] = None,
        planet_file: Optional[str] = None,
    ) -> PlanetData:
        planet_file = _parse_planet_file(
            planet_number=planet_number, planet_file=planet_file
        )
        file = self.directory / planet_file
        return self._loader.load_planet_data(file)

    def _get_ind_output_number(self, time) -> int:
        ini = self._loader.load_ini_file()
        target_time = ini.output_time_interval * self.on
        return find_nearest(time, target_time)

    def find_rp(
        self,
        planet_number: Optional[int] = None,
        *,
        planet_file: Optional[str] = None,
    ) -> float:
        pd = self._load_planet(planet_number=planet_number, planet_file=planet_file)
        ind_on = self._get_ind_output_number(pd.t)
        return pd.d[ind_on]  # type: ignore [attr-defined]

    def find_rhill(
        self,
        planet_number: Optional[int] = None,
        *,
        planet_file: Optional[str] = None,
    ) -> float:
        ini = self._loader.load_ini_file()
        pd = self._load_planet(planet_number=planet_number, planet_file=planet_file)
        oe = pd.get_orbital_elements(ini.frame)
        ind_on = self._get_ind_output_number(pd.t)
        return pow(pd.q[ind_on] / 3.0, 1.0 / 3.0) * oe.a[ind_on]

    def find_phip(
        self,
        planet_number: Optional[int] = None,
        *,
        planet_file: Optional[str] = None,
    ) -> float:
        pd = self._load_planet(planet_number=planet_number, planet_file=planet_file)
        ind_on = self._get_ind_output_number(pd.t)
        return np.arctan2(pd.y, pd.x)[ind_on] % (2 * np.pi)

    @staticmethod
    def _parse_operation_name(
        *,
        prefix: str,
        default_suffix: str,
        operation_name: Optional[str],
    ) -> str:
        if operation_name == "":
            raise ValueError("operation_name cannot be empty")
        suffix = operation_name or default_suffix
        if prefix:
            return f"{prefix}_{suffix}"
        else:
            return suffix

    def latitudinal_projection(self, theta=None, *, operation_name=None) -> "GasField":
        default_suffix = "latitudinal_projection"
        if theta is not None:
            default_suffix += str(np.pi / 2 - theta)
        operation = self._parse_operation_name(
            prefix=self.operation,
            default_suffix=default_suffix,
            operation_name=operation_name,
        )

        imid = self.find_imid()
        if self.native_geometry == "polar":
            ret_coords = Coordinates(
                self.native_geometry,
                self.coords.R,
                self.coords.phi,
                find_around(self.coords.z, self.coords.zmed[imid]),
            )
            # ret_coords = Coordinates(self.native_geometry, self.coords.R, find_around(self.coords.phi, self.coords.phimed[0]), self.coords.z)
            R = self.coords.Rmed
            z = self.coords.zmed
            integral = np.zeros((self.shape[0], self.shape[1]), dtype=">f4")
            # integral = np.zeros((self.shape[0],self.shape[2]), dtype='>f4')
            for i in range(self.shape[0]):
                km = find_nearest(z, z.min())
                kp = find_nearest(z, z.max())
                if theta is not None:
                    km = find_nearest(z, -R[i] * theta)
                    kp = find_nearest(z, R[i] * theta)
                integral[i, :] = np.sum(
                    (self.data[i, :, :] * np.ediff1d(self.coords.z)[None, :])[
                        :, km : kp + 1
                    ],
                    axis=1,
                    dtype="float64",
                )
                # integral[i,km] = -1
                # integral[i,kp] = 1
            ret_data = integral.reshape(self.shape[0], self.shape[1], 1)
            # ret_data = integral.reshape(self.shape[0],1,self.shape[2])
        if self.native_geometry == "spherical":
            ret_coords = Coordinates(
                self.native_geometry,
                self.coords.r,
                find_around(self.coords.theta, self.coords.thetamed[imid]),
                self.coords.phi,
            )
            km = find_nearest(self.coords.theta, self.coords.theta.min())
            kp = find_nearest(self.coords.theta, self.coords.theta.max())
            if theta is not None:
                kp = find_nearest(self.coords.theta, np.pi / 2 + theta)
                km = find_nearest(self.coords.theta, np.pi / 2 - theta)
            ret_data = (
                np.sum(
                    (
                        self.data
                        * self.coords.rmed[:, None, None]
                        * np.sin(self.coords.thetamed[None, :, None])
                        * np.ediff1d(self.coords.theta)[None, :, None]
                    )[:, km : kp + 1, :],
                    axis=1,
                    dtype="float64",
                )
            ).reshape(self.shape[0], 1, self.shape[2])
        return GasField(
            self.field,
            ret_data.astype("float32", copy=False),
            ret_coords,
            self.native_geometry,
            self.on,
            operation,
            inifile=self.inifile,
            directory=self.directory,
            rotate_by=self._rotate_by,
        )

    # def latitudinal_projection(self, theta=None):
    #     operation = self.operation + "_latitudinal_projection"
    #     imid = self.find_imid()
    #     if self.native_geometry == "polar":
    #         ret_coords = Coordinates(
    #             self.native_geometry,
    #             self.coords.R,
    #             self.coords.phi,
    #             find_around(self.coords.z, self.coords.zmed[imid]),
    #         )
    #         # ret_coords = Coordinates(self.native_geometry, self.coords.R, find_around(self.coords.phi, self.coords.phimed[0]), self.coords.z)
    #         R = self.coords.Rmed
    #         z = self.coords.zmed
    #         kall = []
    #         iall = []
    #         integral = np.zeros((self.shape[0], self.shape[1]), dtype=">f4")
    #         for i in range(self.shape[0]):
    #             km = find_nearest(z, z.min())
    #             kp = find_nearest(z, z.max())
    #             if theta is not None:
    #                 km = find_nearest(z, R[i] / np.tan(np.pi / 2 - theta))
    #                 kp = find_nearest(z, R[i] / np.tan(np.pi / 2 + theta))
    #             for k in range(kp, km - 2, -1):
    #                 kall.append(k)
    #                 iall.append(find_nearest(R, np.sqrt(R[i] ** 2 - z[k] ** 2)))
    #             knall = np.array(kall)[
    #                 np.where(
    #                     (z[kall] < R[iall] / np.tan(np.pi / 2 - theta))
    #                     & (z[kall] > R[iall] / np.tan(np.pi / 2 + theta))
    #                 )[0]
    #             ]
    #             inall = np.array(iall)[
    #                 np.where(
    #                     (z[kall] < R[iall] / np.tan(np.pi / 2 - theta))
    #                     & (z[kall] > R[iall] / np.tan(np.pi / 2 + theta))
    #                 )[0]
    #             ]
    #             kall = knall
    #             iall = inall
    #             integral[i, :] = np.nansum(
    #                 self.data[iall[:-1], :, kall[:-1]]
    #                 * R[iall[:-1]][:, None]
    #                 * np.ediff1d(np.arctan2(R[iall], z[kall]))[:, None],
    #                 axis=0,
    #                 dtype="float64",
    #             )
    #             iall = []
    #             kall = []
    #             # integral[i, :] = np.nansum(
    #             #     (self.data[i, :, :] * np.ediff1d(self.coords.z)[None, :])[
    #             #         :, km : kp + 1
    #             #     ],
    #             #     axis=1,
    #             #     dtype="float64",
    #             # )
    #             # integral[i,km] = -1
    #             # integral[i,kp] = 1
    #         ret_data = integral.reshape(self.shape[0], self.shape[1], 1)
    #         # ret_data = integral.reshape(self.shape[0],1,self.shape[2])
    #     if self.native_geometry == "spherical":
    #         ret_coords = Coordinates(
    #             self.native_geometry,
    #             self.coords.r,
    #             find_around(self.coords.theta, self.coords.thetamed[imid]),
    #             self.coords.phi,
    #         )
    #         km = find_nearest(self.coords.thetamed, self.coords.theta.min())
    #         kp = find_nearest(self.coords.thetamed, self.coords.theta.max())
    #         if theta is not None:
    #             km = find_nearest(self.coords.thetamed, np.pi / 2 - theta)
    #             kp = find_nearest(self.coords.thetamed, np.pi / 2 + theta)
    #         ret_data = (
    #             np.nansum(
    #                 (
    #                     self.data
    #                     * self.coords.rmed[:, None, None]
    #                     * np.sin(self.coords.thetamed[None, :, None])
    #                     * np.ediff1d(self.coords.theta)[None, :, None]
    #                 )[:, km : kp + 1, :],
    #                 axis=1,
    #                 dtype="float64",
    #             )
    #         ).reshape(self.shape[0], 1, self.shape[2])
    #     return GasField(
    #         self.field,
    #         np.float32(ret_data),
    #         ret_coords,
    #         self.native_geometry,
    #         self.on,
    #         operation,
    #         inifile=self.inifile,
    #         directory=self.directory,
    #         rotate_by=self._rotate_by,
    #     )

    def vertical_projection(self, z=None, *, operation_name=None) -> "GasField":
        default_suffix = "vertical_projection"
        if z is not None:
            default_suffix += str(z)
        operation = self._parse_operation_name(
            prefix=self.operation,
            default_suffix=default_suffix,
            operation_name=operation_name,
        )

        imid = self.find_imid()
        if self.native_geometry == "cartesian":
            ret_coords = Coordinates(
                self.native_geometry,
                self.coords.x,
                self.coords.y,
                find_around(self.coords.z, self.coords.zmed[imid]),
            )
            km = find_nearest(self.coords.zmed, self.coords.z.min())
            kp = find_nearest(self.coords.zmed, self.coords.z.max())
            if z is not None:
                km = find_nearest(self.coords.zmed, -z)
                kp = find_nearest(self.coords.zmed, z)
            ret_data = (
                np.nansum(
                    (self.data * np.ediff1d(self.coords.z))[:, :, km : kp + 1],
                    axis=2,
                    dtype="float64",
                )
            ).reshape(self.shape[0], self.shape[1], 1)
        if self.native_geometry == "polar":
            ret_coords = Coordinates(
                self.native_geometry,
                self.coords.R,
                self.coords.phi,
                find_around(self.coords.z, self.coords.zmed[imid]),
            )
            km = find_nearest(self.coords.zmed, self.coords.z.min())
            kp = find_nearest(self.coords.zmed, self.coords.z.max())
            if z is not None:
                km = find_nearest(self.coords.zmed, -z)
                kp = find_nearest(self.coords.zmed, z)
            ret_data = (
                np.nansum(
                    (self.data * np.ediff1d(self.coords.z))[:, :, km : kp + 1],
                    axis=2,
                    dtype="float64",
                )
            ).reshape(self.shape[0], self.shape[1], 1)
        if self.native_geometry == "spherical":
            raise NotImplementedError(
                """
                vertical_projection(z) function not implemented in spherical coordinates.\n
                Maybe you could use the function latitudinal_projection(theta)?
                """
            )
        return GasField(
            self.field,
            ret_data.astype("float32", copy=False),
            ret_coords,
            self.native_geometry,
            self.on,
            operation,
            inifile=self.inifile,
            directory=self.directory,
            rotate_by=self._rotate_by,
        )

    def vertical_at_midplane(self, *, operation_name=None) -> "GasField":
        operation = self._parse_operation_name(
            prefix=self.operation,
            default_suffix="vertical_at_midplane",
            operation_name=operation_name,
        )
        imid = self.find_imid()
        if self.native_geometry == "cartesian":
            # find_around looks around the 2 coords values that surround coordmed at imid
            ret_coords = Coordinates(
                self.native_geometry,
                self.coords.x,
                self.coords.y,
                find_around(self.coords.z, self.coords.zmed[imid]),
            )
            # ret_coords = Coordinates(self.native_geometry, self.coords.x, self.coords.y, self.coords.z[imid:imid+2])
            ret_data = self.data[:, :, imid].reshape(self.shape[0], self.shape[1], 1)
            # do geometry conversion!!! -> chainer la conversion (une fois que reduction de dimension -> conversion puis plot egalement chainable)
        if self.native_geometry == "polar":
            ret_coords = Coordinates(
                self.native_geometry,
                self.coords.R,
                self.coords.phi,
                find_around(self.coords.z, self.coords.zmed[imid]),
            )
            # ret_coords = Coordinates(self.native_geometry, self.coords.R, self.coords.phi, self.coords.z[imid:imid+2])
            ret_data = self.data[:, :, imid].reshape(self.shape[0], self.shape[1], 1)
        if self.native_geometry == "spherical":
            ret_coords = Coordinates(
                self.native_geometry,
                self.coords.r,
                find_around(self.coords.theta, self.coords.thetamed[imid]),
                self.coords.phi,
            )
            # ret_coords = Coordinates(self.native_geometry, self.coords.r, self.coords.theta[imid:imid+2], self.coords.phi)
            ret_data = self.data[:, imid, :].reshape(self.shape[0], 1, self.shape[2])
        return GasField(
            self.field,
            ret_data,
            ret_coords,
            self.native_geometry,
            self.on,
            operation,
            inifile=self.inifile,
            directory=self.directory,
            rotate_by=self._rotate_by,
        )

    def latitudinal_at_theta(self, theta=0.0, *, operation_name=None) -> "GasField":
        operation = self._parse_operation_name(
            prefix=self.operation,
            default_suffix=f"latitudinal_at_theta{np.pi/2-theta}",
            operation_name=operation_name,
        )

        imid = self.find_imid(altitude=theta)
        if self.native_geometry == "polar":
            data_at_theta = np.zeros((self.shape[0], self.shape[1]), dtype=">f4")
            for i in range(self.shape[0]):
                if np.sign(theta) >= 0:
                    if (
                        find_nearest(
                            self.coords.zmed,
                            self.coords.R[i] / np.tan(np.pi / 2 - theta),
                        )
                        < self.shape[2]
                    ):
                        # print(i,find_nearest(rhoon.x,rhoon.x[i]),find_nearest(rhoon.z,4*0.05*rhoon.x[i]))
                        # rvz[find_nearest(rhoon.x,rhoon.x[i]),:,find_nearest(rhoon.z,4*0.05*rhoon.x[i])] = 1
                        data_at_theta[i, :] = self.data[
                            i,
                            :,
                            find_nearest(
                                self.coords.zmed,
                                self.coords.R[i] / np.tan(np.pi / 2 - theta),
                            ),
                        ]
                    else:
                        data_at_theta[i, :] = np.nan
                else:
                    if (
                        find_nearest(
                            self.coords.zmed,
                            self.coords.R[i] / np.tan(np.pi / 2 - theta),
                        )
                        > 0
                    ):
                        data_at_theta[i, :] = self.data[
                            i,
                            :,
                            find_nearest(
                                self.coords.zmed,
                                self.coords.R[i] / np.tan(np.pi / 2 - theta),
                            ),
                        ]
                    else:
                        data_at_theta[i, :] = np.nan
            ret_data = data_at_theta.reshape(self.shape[0], self.shape[1], 1)
            ret_coords = Coordinates(
                self.native_geometry,
                self.coords.R,
                self.coords.phi,
                find_around(self.coords.z, self.coords.zmed[imid]),
            )
        if self.native_geometry == "spherical":
            ret_coords = Coordinates(
                self.native_geometry,
                self.coords.r,
                find_around(self.coords.theta, self.coords.thetamed[imid]),
                self.coords.phi,
            )
            ret_data = self.data[
                :, find_nearest(self.coords.thetamed, np.pi / 2 - theta), :
            ].reshape(self.shape[0], 1, self.shape[2])
        return GasField(
            self.field,
            ret_data,
            ret_coords,
            self.native_geometry,
            self.on,
            operation,
            inifile=self.inifile,
            directory=self.directory,
            rotate_by=self._rotate_by,
        )

    def vertical_at_z(self, z=0.0, *, operation_name=None) -> "GasField":
        logger.info("vertical_at_z TO BE TESTED")
        operation = self._parse_operation_name(
            prefix=self.operation,
            default_suffix=f"vertical_at_z{z}",
            operation_name=operation_name,
        )
        imid = self.find_imid(altitude=z)
        if self.native_geometry == "cartesian":
            ret_coords = Coordinates(
                self.native_geometry,
                self.coords.x,
                self.coords.y,
                find_around(self.coords.z, self.coords.zmed[imid]),
            )
            ret_data = self.data[:, :, find_nearest(self.coords.zmed, z)].reshape(
                self.shape[0], self.shape[1], 1
            )
        if self.native_geometry == "polar":
            ret_coords = Coordinates(
                self.native_geometry,
                self.coords.R,
                self.coords.phi,
                find_around(self.coords.z, self.coords.zmed[imid]),
            )
            ret_data = self.data[:, :, find_nearest(self.coords.zmed, z)].reshape(
                self.shape[0], self.shape[1], 1
            )
        if self.native_geometry == "spherical":
            raise NotImplementedError(
                "vertical at z in spherical coordinates not implemented yet."
            )
        """
            data_at_z = np.zeros((self.shape[0], self.shape[2]), dtype=">f4")
            for i in range(self.shape[0]):
                if np.sign(z) >= 0:
                    if (
                        find_nearest(
                            self.coords.thetamed, np.arctan2(self.coords.r[i], z)
                        )
                        < self.shape[2]
                    ):
                        # print(i,find_nearest(rhoon.x,rhoon.x[i]),find_nearest(rhoon.z,4*0.05*rhoon.x[i]))
                        # rvz[find_nearest(rhoon.x,rhoon.x[i]),:,find_nearest(rhoon.z,4*0.05*rhoon.x[i])] = 1
                        data_at_z[i, :] = self.data[
                            i,
                            find_nearest(
                                self.coords.thetamed, np.arctan2(self.coords.r[i], z)
                            ),
                            :,
                        ]
                    else:
                        data_at_z[i, :] = np.nan
                else:
                    if (
                        find_nearest(
                            self.coords.thetamed, np.arctan2(self.coords.r[i], z)
                        )
                        > 0
                    ):
                        data_at_z[i, :] = self.data[
                            i,
                            find_nearest(
                                self.coords.thetamed, np.arctan2(self.coords.r[i], z)
                            ),
                            :,
                        ]
                    else:
                        data_at_z[i, :] = np.nan
            ret_data = data_at_z.reshape(self.shape[0], 1, self.shape[2])
            ret_coords = Coordinates(
                self.native_geometry,
                self.coords.r,
                find_around(self.coords.theta, self.coords.thetamed[imid]),
                self.coords.phi,
            )
        """
        return GasField(
            self.field,
            ret_data,
            ret_coords,
            self.native_geometry,
            self.on,
            operation,
            inifile=self.inifile,
            directory=self.directory,
            rotate_by=self._rotate_by,
        )

    def azimuthal_at_phi(self, phi=0.0, *, operation_name=None) -> "GasField":
        operation = self._parse_operation_name(
            prefix=self.operation,
            default_suffix=f"azimuthal_at_phi{phi}",
            operation_name=operation_name,
        )
        iphi = self.find_iphi(phi=phi)
        if self.native_geometry == "cartesian":
            raise NotImplementedError(
                f"geometry flag '{self.native_geometry}' not implemented yet for azimuthal_at_phi"
            )
        if self.native_geometry == "polar":
            ret_coords = Coordinates(
                self.native_geometry,
                self.coords.R,
                find_around(self.coords.phi, self.coords.phimed[iphi]),
                self.coords.z,
            )
            ret_data = self.data[:, iphi, :].reshape(self.shape[0], 1, self.shape[2])
        if self.native_geometry == "spherical":
            ret_coords = Coordinates(
                self.native_geometry,
                self.coords.r,
                self.coords.theta,
                find_around(self.coords.phi, self.coords.phimed[iphi]),
            )
            ret_data = self.data[:, :, iphi].reshape(self.shape[0], self.shape[1], 1)
        return GasField(
            self.field,
            ret_data,
            ret_coords,
            self.native_geometry,
            self.on,
            operation,
            inifile=self.inifile,
            directory=self.directory,
            rotate_by=self._rotate_by,
        )

    def azimuthal_at_planet(
        self,
        planet_number: Optional[int] = None,
        *,
        planet_file: Optional[str] = None,
        operation_name=None,
    ) -> "GasField":
        planet_file = _parse_planet_file(
            planet_number=planet_number, planet_file=planet_file
        )
        del planet_number

        operation = self._parse_operation_name(
            prefix=self.operation,
            default_suffix="azimuthal_at_planet",
            operation_name=operation_name,
        )

        phip = self.find_phip(planet_file=planet_file)
        aziphip = self.azimuthal_at_phi(phi=phip)
        return GasField(
            self.field,
            aziphip.data,
            aziphip.coords,
            self.native_geometry,
            self.on,
            operation,
            inifile=self.inifile,
            directory=self.directory,
            rotate_by=self._rotate_by,
        )

    def azimuthal_average(self, *, operation_name=None) -> "GasField":
        operation = self._parse_operation_name(
            prefix=self.operation,
            default_suffix="azimuthal_average",
            operation_name=operation_name,
        )

        iphi = self.find_iphi(phi=0)
        if self.native_geometry == "cartesian":
            raise NotImplementedError(
                f"geometry flag '{self.native_geometry}' not implemented yet for azimuthal_average"
            )
        if self.native_geometry == "polar":
            ret_coords = Coordinates(
                self.native_geometry,
                self.coords.R,
                find_around(self.coords.phi, self.coords.phimed[iphi]),
                self.coords.z,
            )
            ret_data = np.nanmean(self.data, axis=1, dtype="float64").reshape(
                self.shape[0], 1, self.shape[2]
            )
        if self.native_geometry == "spherical":
            ret_coords = Coordinates(
                self.native_geometry,
                self.coords.r,
                self.coords.theta,
                find_around(self.coords.phi, self.coords.phimed[iphi]),
            )
            ret_data = np.nanmean(self.data, axis=2, dtype="float64").reshape(
                self.shape[0], self.shape[1], 1
            )
        return GasField(
            self.field,
            ret_data.astype("float32", copy=False),
            ret_coords,
            self.native_geometry,
            self.on,
            operation,
            inifile=self.inifile,
            directory=self.directory,
            rotate_by=self._rotate_by,
        )

    def remove_planet_hill_band(
        self,
        planet_number: Optional[int] = None,
        *,
        planet_file: Optional[str] = None,
        operation_name=None,
    ) -> "GasField":
        planet_file = _parse_planet_file(
            planet_number=planet_number, planet_file=planet_file
        )
        del planet_number

        operation = self._parse_operation_name(
            prefix=self.operation,
            default_suffix="remove_planet_hill_band",
            operation_name=operation_name,
        )

        phip = self.find_phip(planet_file=planet_file)
        rp = self.find_rp(planet_file=planet_file)
        rhill = self.find_rhill(planet_file=planet_file)
        iphip_m = self.find_iphi(phi=phip - 2 * rhill / rp)
        iphip_p = self.find_iphi(phi=phip + 2 * rhill / rp)
        if self.native_geometry == "cartesian":
            raise NotImplementedError(
                f"geometry flag '{self.native_geometry}' not implemented yet for azimuthal_average_except_planet_hill"
            )
        if self.native_geometry == "polar":
            ret_coords = Coordinates(
                self.native_geometry,
                self.coords.R,
                # find_around(self.coords.phi, phip),
                self.coords.phi,
                self.coords.z,
            )
            ret_data = self.data.copy()
            if iphip_p >= iphip_m and iphip_p != self.coords.shape[1]:
                ret_data[:, iphip_m : iphip_p + 1, :] = np.nan
            else:
                if iphip_p == self.coords.shape[1]:
                    ret_data[:, iphip_m:iphip_p, :] = np.nan
                else:
                    ret_data[:, 0 : iphip_p + 1, :] = np.nan
                    ret_data[:, iphip_m : self.coords.shape[1], :] = np.nan
            # ret_data = np.nanmean(self.data, axis=1, dtype="float64").reshape(
            #     self.shape[0], 1, self.shape[2]
            # )
        if self.native_geometry == "spherical":
            ret_coords = Coordinates(
                self.native_geometry,
                self.coords.r,
                self.coords.theta,
                # find_around(self.coords.phi, phip),
                self.coords.phi,
            )
            ret_data = self.data.copy()
            if iphip_p >= iphip_m and iphip_p != self.coords.shape[2]:
                ret_data[:, :, iphip_m : iphip_p + 1] = np.nan
            else:
                if iphip_p == self.coords.shape[2]:
                    ret_data[:, :, iphip_m:iphip_p] = np.nan
                else:
                    ret_data[:, :, 0 : iphip_p + 1] = np.nan
                    ret_data[:, :, iphip_m : self.coords.shape[2]] = np.nan
            # ret_data = np.nanmean(self.data, axis=2, dtype="float64").reshape(
            #     self.shape[0], self.shape[1], 1
            # )
        return GasField(
            self.field,
            ret_data,
            ret_coords,
            self.native_geometry,
            self.on,
            operation,
            inifile=self.inifile,
            directory=self.directory,
            rotate_by=self._rotate_by,
        )

    def radial_at_r(self, distance=1.0, *, operation_name=None) -> "GasField":
        operation = self._parse_operation_name(
            prefix=self.operation,
            default_suffix=f"radial_at_r{distance}",
            operation_name=operation_name,
        )

        ir1 = self.find_ir(distance=distance)
        if self.native_geometry == "cartesian":
            raise NotImplementedError(
                f"geometry flag '{self.native_geometry}' not implemented yet for radial_at_r"
            )
        if self.native_geometry == "polar":
            # self.field = r"%s (R=1)" % self.field
            ret_coords = Coordinates(
                self.native_geometry,
                find_around(self.coords.R, self.coords.Rmed[ir1]),
                self.coords.phi,
                self.coords.z,
            )
        if self.native_geometry == "spherical":
            # self.field = r"%s (r=1)" % self.field
            ret_coords = Coordinates(
                self.native_geometry,
                find_around(self.coords.r, self.coords.rmed[ir1]),
                self.coords.theta,
                self.coords.phi,
            )
        ret_data = self.data[ir1, :, :].reshape(1, self.shape[1], self.shape[2])
        return GasField(
            self.field,
            ret_data,
            ret_coords,
            self.native_geometry,
            self.on,
            operation,
            inifile=self.inifile,
            directory=self.directory,
            rotate_by=self._rotate_by,
        )

    def radial_average_interval(
        self, vmin=None, vmax=None, *, operation_name=None
    ) -> "GasField":
        if (vmin is None) or (vmax is None):
            raise ValueError(
                f"The radial interval {vmin=} and {vmax=} should be defined"
            )

        operation = self._parse_operation_name(
            prefix=self.operation,
            default_suffix=f"radial_average_interval_{vmin}_{vmax}",
            operation_name=operation_name,
        )

        irmin = self.find_ir(distance=vmin)
        irmax = self.find_ir(distance=vmax)
        ir = self.find_ir(distance=(vmax - vmin) / 2)
        if self.native_geometry == "cartesian":
            raise NotImplementedError(
                f"geometry flag '{self.native_geometry}' not implemented yet for radial_at_r"
            )
        if self.native_geometry == "polar":
            if vmin is None:
                vmin = self.coords.R.min()
            if vmax is None:
                vmax = self.coords.R.max()
            ret_coords = Coordinates(
                self.native_geometry,
                find_around(self.coords.R, self.coords.Rmed[ir]),
                self.coords.phi,
                self.coords.z,
            )
        if self.native_geometry == "spherical":
            if vmin is None:
                vmin = self.coords.r.min()
            if vmax is None:
                vmax = self.coords.r.max()
            ret_coords = Coordinates(
                self.native_geometry,
                find_around(self.coords.r, self.coords.rmed[ir]),
                self.coords.theta,
                self.coords.phi,
            )
        ret_data = np.nanmean(
            self.data[irmin : irmax + 1, :, :], axis=0, dtype="float64"
        ).reshape(1, self.shape[1], self.shape[2])
        return GasField(
            self.field,
            ret_data,
            ret_coords,
            self.native_geometry,
            self.on,
            operation,
            inifile=self.inifile,
            directory=self.directory,
            rotate_by=self._rotate_by,
        )

    def diff(self, on_2) -> "GasField":
        ds_2 = GasDataSet(
            on_2,
            geometry=self.native_geometry,
            inifile=self.inifile,
            directory=self.directory,
        )
        if self.operation != "":
            raise KeyError(
                "For now, diff should only be applied on the initial Field cube."
            )
        # self.operation += "diff"
        ret_data = (self.data - ds_2[self.field].data) / ds_2[self.field].data
        ret_coords = self.coords
        # self.field = r"$\frac{%s - %s_0}{%s_0}$" % (self.field, self.field, self.field)
        return GasField(
            self.field,
            ret_data,
            ret_coords,
            self.native_geometry,
            self.on,
            self.operation,
            inifile=self.inifile,
            directory=self.directory,
            rotate_by=self._rotate_by,
        )

    def rotate(
        self,
        planet_corotation: Optional[int] = None,
        *,
        rotate_with: Optional[str] = None,
        rotate_by: Optional[float] = None,
    ) -> "GasField":
        rotate_by = _parse_rotation_angle(
            rotate_by=rotate_by,
            rotate_with=rotate_with,
            planet_number_argument=("planet_corotation", planet_corotation),
            find_phip=self.find_phip,
            stacklevel=2,
        )

        operation = self.operation
        if self.shape.count(1) > 1:
            raise ValueError("data has to be 2D or 3D in order to rotate the data.")
        if not _fequal(self._rotate_by, rotate_by):
            phicoord = self.coords.phi - rotate_by
            ipi = find_nearest(phicoord, 0)
            if abs(0 - phicoord[ipi]) > abs(np.ediff1d(find_around(phicoord, 0))[0]):
                ipi = find_nearest(phicoord, 2 * np.pi)
            if self.native_geometry == "polar":
                ret_data = np.roll(
                    self.data,
                    -ipi + 1,
                    axis=1,
                )
                ret_coords = Coordinates(
                    self.native_geometry, self.coords.R, self.coords.phi, self.coords.z
                )
            elif self.native_geometry == "spherical":
                ret_data = np.roll(
                    self.data,
                    -ipi + 1,
                    axis=2,
                )
                ret_coords = Coordinates(
                    self.native_geometry,
                    self.coords.r,
                    self.coords.theta,
                    self.coords.phi,
                )
            else:
                raise NotImplementedError(
                    f"geometry flag '{self.native_geometry}' not implemented yet if corotation"
                )
            self._rotate_by = rotate_by
        else:
            ret_data = self.data
            if self.native_geometry == "polar":
                ret_coords = Coordinates(
                    self.native_geometry, self.coords.R, self.coords.phi, self.coords.z
                )
            elif self.native_geometry == "spherical":
                ret_coords = Coordinates(
                    self.native_geometry,
                    self.coords.r,
                    self.coords.theta,
                    self.coords.phi,
                )

        return GasField(
            self.field,
            ret_data,
            ret_coords,
            self.native_geometry,
            self.on,
            operation,
            inifile=self.inifile,
            directory=self.directory,
            rotate_by=self._rotate_by,
        )


class GasDataSet:
    """Idefix dataset class that contains everything in the .vtk file

    Args:
        input_dataset (int or str): output number or file name
        directory (str): directory of the .vtk
        geometry (str): for retrocompatibility if old vtk format
        inifile (str): name of the simulation's parameter file if no default files (combined with code)
        code (str): name of the code ("idefix", "pluto", "fargo3d", "fargo-adsg")
    Returns:
        dataset
    """

    def __init__(
        self,
        input_dataset: Union[int, PathT],
        /,
        *,
        inifile: Optional[PathT] = None,
        code: Union[str, Recipe, None] = None,
        geometry: Optional[str] = None,
        directory: Optional[PathT] = None,
        fluid: Optional[str] = None,
        operation: Optional[str] = None,
    ) -> None:
        if isinstance(input_dataset, (str, Path)):
            input_dataset = Path(input_dataset)
            directory_from_input = input_dataset.parent
            if directory is None:
                directory = directory_from_input
            elif directory_from_input.resolve() != Path(directory).resolve():
                raise ValueError(
                    f"directory value {directory!r} does not match "
                    f"directory name from input_dataset ({directory_from_input!r})"
                )
            del directory_from_input

        if directory is None:
            directory = Path.cwd()
        else:
            directory = Path(directory)

        recipe = recipe_from(
            code=code,
            parameter_file=inifile,
            directory=directory,
        )

        if fluid is not None and recipe is not Recipe.FARGO3D:
            warnings.warn(
                "Unused keyword argument: 'fluid'",
                category=UserWarning,
                stacklevel=2,
            )

        loader = loader_from(
            code=code,
            parameter_file=inifile,
            directory=directory,
        )
        if operation is not None:
            ignored_kwargs = []
            msg = ""
            if fluid is not None:
                ignored_kwargs.append("fluid")
            if geometry is not None:
                ignored_kwargs.append("geometry")
            if ignored_kwargs:
                ignored = ", ".join(repr(_) for _ in ignored_kwargs)
                msg = (
                    "The following keyword arguments are ignored "
                    f"when combined with 'operation': {ignored}"
                )
                warnings.warn(msg, UserWarning, stacklevel=2)
            self._loader = dataclasses.replace(loader, binary_reader=NPYReader)
        else:
            self._loader = loader

        self.on, datafile = self._loader.binary_reader.parse_output_number_and_filename(
            input_dataset,
            directory=directory,
            prefix=operation or "",
        )

        self._read = self._loader.load_bin_data(
            datafile,
            geometry=geometry,
            fluid=fluid,
        )

        self.native_geometry = self._read.geometry
        self.dict = self._read.data
        self.coords = Coordinates(
            self.native_geometry,
            self._read.x1,
            self._read.x2,
            self._read.x3,
        )
        for key in self.dict:
            self.dict[key] = GasField(
                key,
                self.dict[key],
                self.coords,
                self.native_geometry,
                self.on,
                operation="",
                inifile=self._loader.parameter_file,
                code=recipe,
                directory=directory,
            )

        # backward compatibility for self.params
        self._parameters_input = {
            "inifile": inifile,
            "code": code.removesuffix("_vtk") if code is not None else None,
            "directory": directory,
        }

    @cached_property
    def params(self):
        from nonos.api.from_simulation import Parameters

        warnings.warn(
            "GasDataSet.params is deprecated "
            "and will be removed in a future version.",
            category=DeprecationWarning,
            stacklevel=2,
        )
        return Parameters(**self._parameters_input)

    @classmethod
    def from_npy(
        cls,
        on: int,
        *,
        inifile: Optional[PathT] = None,
        code: Union[str, Recipe, None] = None,
        directory: Optional[PathT] = None,
        operation: str,
    ) -> "GasDataSet":
        warnings.warn(
            "GasDataSet.from_npy is deprecated "
            "and will be removed in a future version. "
            "Instead, call GasDataSet(...) directly.",
            category=DeprecationWarning,
            stacklevel=2,
        )
        return GasDataSet(
            on,
            inifile=inifile,
            code=code,
            directory=directory,
            operation=operation,
        )

    def __getitem__(self, key) -> "GasField":
        if key in self.dict:
            return self.dict[key]
        else:
            raise KeyError

    def keys(self) -> KeysView[str]:
        """
        Returns
        =======
        keys of the dict
        """
        return self.dict.keys()

    def values(self) -> ValuesView["GasField"]:
        """
        Returns
        =======
        values of the dict
        """
        return self.dict.values()

    def items(self) -> ItemsView[str, "GasField"]:
        """
        Returns
        =======
        items of the dict
        """
        return self.dict.items()

    @property
    def nfields(self) -> int:
        """
        Returns
        =======
        The number of fields in the GasDataSet
        """
        return len(self.dict)



import sys
from typing import Union

from loguru import logger
from rich import print as rprint
from rich.logging import RichHandler


def configure_logger(level: Union[int, str] = 30, **kwargs) -> None:
    logger.remove()  # remove pre-existing handler
    logger.add(
        RichHandler(
            log_time_format="[%X] nonos",
            omit_repeated_times=False,
        ),
        format="{message}",
        level=level,
        **kwargs,
    )


def print_warn(message) -> None:
    """
    adapted from idefix_cli (cmt robert)
    https://github.com/neutrinoceros/idefix_cli
    """
    rprint(f":bone: [bold red]Warning[/] {message}", file=sys.stderr)


def print_err(message) -> None:
    """
    adapted from idefix_cli (cmt robert)
    https://github.com/neutrinoceros/idefix_cli
    """
    rprint(f":bone: [bold white on red]Error[/] {message}", file=sys.stderr)


def parse_verbose_level(verbose: int) -> str:
    levels = ["WARNING", "INFO", "DEBUG"]
    level = levels[min(len(levels) - 1, verbose)]  # capped to number of levels
    return level


configure_logger(level=30)



__all__ = [
    "NullReader",
    "IdefixReader",
    "Fargo3DReader",
    "FargoADSGReader",
]
import re
from abc import ABC, abstractmethod
from pathlib import Path
from typing import final

import numpy as np

from nonos._readers._base import ReaderMixin
from nonos._types import PathT, PlanetData


@final
class NullReader(ReaderMixin):
    @staticmethod
    def get_planet_files(directory: Path, /) -> list[Path]:
        raise NotImplementedError(
            f"{directory} couldn't be read. The default reader class (NullReader) "
            "was previously selected, possibly by mistake ?"
        )

    @staticmethod
    def read(file: PathT, /) -> PlanetData:
        raise NotImplementedError(
            f"{file} couldn't be read. The default reader class (NullReader) "
            "was previously selected, possibly by mistake ?"
        )


@final
class IdefixReader(ReaderMixin):
    @staticmethod
    def get_planet_files(directory: Path, /) -> list[Path]:
        return sorted(directory.glob("planet*.dat"))

    @staticmethod
    def read(file: PathT, /) -> PlanetData:
        dt, x, y, z, vx, vy, vz, q, t = np.loadtxt(file).T
        return PlanetData(x, y, z, vx, vy, vz, q, t, dt)


class FargoReader(ReaderMixin, ABC):
    @staticmethod
    def get_planet_files(directory: Path, /) -> list[Path]:
        return [
            fn
            for fn in sorted(directory.glob("planet*.dat"))
            if re.search(r"planet\d+.dat$", str(fn)) is not None
        ]

    @staticmethod
    @abstractmethod
    def read(file: PathT, /) -> PlanetData: ...


@final
class Fargo3DReader(FargoReader):
    @staticmethod
    def read(file: PathT, /) -> PlanetData:
        dt, x, y, z, vx, vy, vz, q, t, *_ = np.loadtxt(file).T
        return PlanetData(x, y, z, vx, vy, vz, q, t, dt)


@final
class FargoADSGReader(FargoReader):
    @staticmethod
    def read(file: PathT, /) -> PlanetData:
        dt, x, y, vx, vy, q, _, _, t, *_ = np.loadtxt(file).T
        z = np.zeros_like(x)
        vz = np.zeros_like(vx)
        return PlanetData(x, y, z, vx, vy, vz, q, t, dt)



DEFAULTS = {
    # directory where the simulation results are stored
    "datadir": ".",
    # which field?
    "field": "RHO",
    # which operation?
    "operation": "unset",
    # if latitudinal operation
    "theta": "unset",
    # if vertical operation
    "z": "unset",
    # if azimuthal operation
    "phi": "unset",
    # if radial operation
    "distance": "unset",
    # if the geometry of idefix outputs is not recognized
    "geometry": "unset",
    # a single output number ('on') to display or a range (min, max, [step])
    "on": "unset",
    # perturbation of field
    "diff": False,
    # field in log
    "log": False,
    # default extent of the matplotlib window
    "range": "unset",
    # default min field
    "vmin": "unset",
    # default max field
    "vmax": "unset",
    # plane to represent (default is midplane)
    "plane": "unset",
    # do we display the progress (loading+plotting)
    "progressBar": False,
    # do the grid rotate with the planet?
    "corotate": "unset",
    # number of cpus to use
    "ncpu": 1,
    # scaling factor for font size of text in graphs (among other things)
    "scaling": 1,
    # choice of colormap
    "cmap": "RdYlBu_r",
    # name of colorbar
    "title": "unset",
    # conversion factor
    "unit_conversion": 1,
    # select image file format
    "format": "unset",
    # select image resolution
    "dpi": 200,
}



import numpy as np


def find_nearest(array, value):
    array = np.asarray(array)
    idx = (np.abs(array - value)).argmin()
    return idx


def find_around(array, value):
    array = np.asarray(array)
    idx_1 = (np.abs(array - value)).argmin()
    larray = list(array)
    larray.remove(larray[idx_1])
    arraym = np.asarray(larray)
    idx_2 = (np.abs(arraym - value)).argmin()
    return np.asarray([array[idx_1], arraym[idx_2]])



import os

import inifix
import pytest

from nonos.config import DEFAULTS
from nonos.main import main


@pytest.fixture()
def minimal_paramfile(tmp_path):
    ifile = tmp_path / "nonos.ini"
    # check that this setup still makes sense
    assert DEFAULTS["field"] == "RHO"
    with open(ifile, "w") as fh:
        fh.write("field  VX1")
    return ifile


def test_load_config_file(minimal_paramfile, capsys):
    os.chdir(minimal_paramfile.parent)
    ret = main(["-input", "nonos.ini", "-config"])

    assert ret == 0
    out, err = capsys.readouterr()
    assert "Using parameters from" in err
    conf = inifix.loads(out)
    assert conf["field"] == "VX1"


def test_isolated_mode(minimal_paramfile, capsys):
    os.chdir(minimal_paramfile.parent)
    ret = main(["-config", "-isolated"])

    assert ret == 0
    out, err = capsys.readouterr()
    assert err == ""
    conf = inifix.loads(out)
    assert conf["field"] == DEFAULTS["field"]



import os
import shutil

import numpy as np
import pytest

from nonos.api import GasDataSet


def test_from_npy_error(test_data_dir):
    with pytest.raises(FileNotFoundError):
        GasDataSet(
            500,
            operation="typo",
            directory=test_data_dir / "idefix_spherical_planet3d",
        )


def test_roundtrip_simple(test_data_dir, tmp_path):
    ds = GasDataSet(500, directory=test_data_dir / "idefix_spherical_planet3d")
    assert ds.nfields == 7

    gf = ds["RHO"].azimuthal_average()

    gf.save(tmp_path)
    dsnpy = GasDataSet(
        500,
        operation="azimuthal_average",
        directory=tmp_path,
    )
    assert dsnpy.nfields == 1


@pytest.mark.parametrize(
    "implicit_directory",
    [
        pytest.param(True, id="implicit directory"),
        pytest.param(False, id="explicit directory"),
    ],
)
def test_simple_fargo_adsg(test_data_dir, implicit_directory):
    data_dir = test_data_dir / "fargo_adsg_planet"
    if implicit_directory:
        os.chdir(data_dir)
        directory = None
    else:
        directory = data_dir.absolute()

    ds = GasDataSet(
        200,
        code="fargo_adsg",
        inifile="planetpendragon_200k.par",
        directory=directory,
    )
    assert ds.nfields == 1


def test_roundtrip_no_operation_all_field(test_data_dir, tmp_path):
    os.chdir(test_data_dir / "idefix_spherical_planet3d")

    ds = GasDataSet(500)
    assert ds.nfields == 7

    gf = ds["RHO"]

    gf.save(tmp_path)
    dsnpy = GasDataSet(
        500,
        operation="",
        directory=tmp_path,
    )
    assert dsnpy.nfields == 1
    np.testing.assert_array_equal(ds["RHO"].data, dsnpy["RHO"].data)


def test_roundtrip_other_dir(test_data_dir, tmp_path):
    os.chdir(test_data_dir / "idefix_spherical_planet3d")
    gf = GasDataSet(500)["RHO"].azimuthal_average()
    gf.save(tmp_path)
    dsnpy = GasDataSet(
        500,
        operation="azimuthal_average",
        directory=tmp_path,
    )
    assert dsnpy.nfields == 1


@pytest.mark.parametrize("header_only", [True, False])
def test_save_to_new_dir(header_only, test_data_dir, tmp_path):
    gf = GasDataSet(500, directory=test_data_dir / "idefix_spherical_planet3d")["RHO"]

    savedir = tmp_path / "savedir"

    # exercise saving to a directory that doesn't exist yet
    saved_file = gf.save(savedir, header_only=header_only)
    assert savedir.exists
    assert savedir == saved_file.parents[1]
    if header_only:
        assert not saved_file.exists()
    else:
        assert saved_file.is_file()

        # exercise saving again
        write_time = saved_file.stat().st_mtime
        saved_file_2 = gf.save(savedir, header_only=header_only)
        assert saved_file_2 == saved_file
        assert saved_file_2.stat().st_mtime == write_time, "File was overwritten"


def test_npy_radial_at_r(test_data_dir, tmp_path):
    os.chdir(test_data_dir / "idefix_spherical_planet3d")
    gf = GasDataSet(500)["RHO"].radial_at_r(1.1)
    gf.save(tmp_path)
    dsnpy = GasDataSet(
        500,
        operation="radial_at_r1.1",
        directory=tmp_path,
    )
    assert list(dsnpy.keys()) == ["RHO"]


def test_latitudinal_projection(test_data_dir):
    os.chdir(test_data_dir / "idefix_spherical_planet3d")
    theta = np.arctan(5 * 0.1)
    gflpt = GasDataSet(500)["RHO"].latitudinal_projection(theta)
    gflp = GasDataSet(500)["RHO"].latitudinal_projection()
    np.testing.assert_equal(gflpt.data, gflp.data)


def test_radial_average_interval_vmin_vmax(test_data_dir):
    os.chdir(test_data_dir / "idefix_spherical_planet3d")
    with pytest.raises(
        ValueError,
        match=r"The radial interval vmin=1 and vmax=None should be defined",
    ):
        GasDataSet(500)["RHO"].radial_average_interval(1)


def test_default_operation_name(test_data_dir):
    os.chdir(test_data_dir / "idefix_spherical_planet3d")
    gf = GasDataSet(500)["RHO"].radial_average_interval(1, 2)
    assert gf.operation == "radial_average_interval_1_2"


def test_custom_operation_name(test_data_dir):
    os.chdir(test_data_dir / "idefix_spherical_planet3d")
    gfno = GasDataSet(500)["RHO"].radial_average_interval(
        1, 2, operation_name="radial_custom"
    )
    assert gfno.operation == "radial_custom"


def test_save_current_dir(test_data_dir, tmp_path):
    src_dir = test_data_dir / "idefix_spherical_planet3d"
    shutil.copy(src_dir / "idefix.ini", tmp_path)
    shutil.copy(src_dir / "data.0500.vtk", tmp_path)

    inifile = tmp_path / "idefix.ini"
    time_inifile = inifile.stat().st_mtime
    gf = GasDataSet(500, directory=tmp_path)["RHO"]
    gf.save(tmp_path)
    assert inifile.stat().st_mtime == time_inifile


@pytest.mark.parametrize(
    "from_abs_path",
    [
        pytest.param(True, id="from absolute path"),
        pytest.param(False, id="from relative path"),
    ],
)
def test_api_vtk_by_name(test_data_dir, from_abs_path):
    data_dir = test_data_dir / "idefix_spherical_planet3d"
    if from_abs_path:
        input_ = str((data_dir / "data.0500.vtk").absolute())
    else:
        os.chdir(data_dir)
        input_ = "data.0500.vtk"

    ds = GasDataSet(input_)
    assert ds.on == 500

    with pytest.raises(FileNotFoundError):
        GasDataSet(input_.replace("data.0500", "datawrong.0500"))


def test_api_vtk_by_name_fargo(test_data_dir):
    GasDataSet(test_data_dir / "fargo3d_planet2d" / "gasdens40.dat")


def test_api_fluid_fargo3d(test_data_dir):
    args = (5,)
    kwargs = {
        "fluid": "dust2",
        "directory": test_data_dir / "fargo3d_multifluid",
    }
    ds = GasDataSet(*args, **kwargs)
    assert ds.nfields == 1

    kwargs["fluid"] = "dust999"
    with pytest.raises(
        FileNotFoundError,
        match=r"No file matches the pattern 'dust999\*5\.dat'",
    ):
        GasDataSet(*args, **kwargs)


def test_api_fluid_idefix(test_data_dir):
    with pytest.warns(
        UserWarning,
        match="Unused keyword argument: 'fluid'",
    ):
        GasDataSet(
            500,
            fluid="dust1",
            directory=test_data_dir / "idefix_spherical_planet3d",
        )


# fmt: off
@pytest.mark.parametrize(
    "geometry, slice_no, operation_name, operation_args, axis",
    [
        pytest.param("spherical_3d", 1, "azimuthal_at_phi", (np.pi,), "r"),
        pytest.param("spherical_3d", 1, "azimuthal_at_phi", (np.pi,), "theta"),
        pytest.param("spherical_3d", 1, "azimuthal_at_phi", (np.pi,), "phi", marks=pytest.mark.xfail(strict=True, reason="known bug in azimuthal_at_phi")),
        pytest.param("spherical_3d", 2, "vertical_at_midplane", (), "r"),
        pytest.param("spherical_3d", 2, "vertical_at_midplane", (), "theta", marks=pytest.mark.xfail(strict=True, reason="known bug in vertical_at_midplane")),
        pytest.param("spherical_3d", 2, "vertical_at_midplane", (), "phi"),
        pytest.param("spherical_3d", 3, "radial_at_r", (1.0,), "r", marks=pytest.mark.xfail(strict=True, reason="known bug in radial_at_r")),
        pytest.param("spherical_3d", 3, "radial_at_r", (1.0,), "theta"),
        pytest.param("spherical_3d", 3, "radial_at_r", (1.0,), "phi"),
        pytest.param("spherical_3d", 4, "azimuthal_average", (), "r"),
        pytest.param("spherical_3d", 4, "azimuthal_average", (), "theta"),
        pytest.param("spherical_3d", 4, "azimuthal_average", (), "phi", marks=pytest.mark.xfail(strict=True, reason="known bug in azimuthal_average")),

        pytest.param("polar_3d", 1, "azimuthal_at_phi", (np.pi,), "R"),
        pytest.param("polar_3d", 1, "azimuthal_at_phi", (np.pi,), "phi", marks=pytest.mark.xfail(strict=True, reason="known bug in azimuthal_at_phi")),
        pytest.param("polar_3d", 1, "azimuthal_at_phi", (np.pi,), "z"),
        pytest.param("polar_3d", 2, "vertical_at_midplane", (), "R"),
        pytest.param("polar_3d", 2, "vertical_at_midplane", (), "phi"),
        pytest.param("polar_3d", 2, "vertical_at_midplane", (), "z", marks=pytest.mark.xfail(strict=True, reason="known bug in vertical_at_midplane")),
        pytest.param("polar_3d", 3, "radial_at_r", (2.0078125,), "R", marks=pytest.mark.xfail(strict=True, reason="known bug in radial_at_r")),
        pytest.param("polar_3d", 3, "radial_at_r", (2.0078125,), "phi"),
        pytest.param("polar_3d", 3, "radial_at_r", (2.0078125,), "z"),
        pytest.param("polar_3d", 4, "azimuthal_average", (), "R"),
        pytest.param("polar_3d", 4, "azimuthal_average", (), "phi", marks=pytest.mark.xfail(strict=True, reason="known bug in azimuthal_average")),
        pytest.param("polar_3d", 4, "azimuthal_average", (), "z"),
    ],
)
# fmt: on
def test_api_vtk_slices_idefix(test_data_dir, geometry, slice_no, operation_name, operation_args, axis):
    on = 9
    ds = GasDataSet(test_data_dir / "idefix_vtk_slices" / geometry / f"data.{on:04d}.vtk")
    ds_phi_cut = GasDataSet(
        test_data_dir / "idefix_vtk_slices" / geometry / f"slice{slice_no}.{on:04d}.vtk"
    )
    method = getattr(ds["RHO"], operation_name)
    rho_slice_pp = method(*operation_args)
    rho_slice_cut = ds_phi_cut["RHO"]
    np.testing.assert_array_almost_equal(rho_slice_pp.data, rho_slice_cut.data)

    np.testing.assert_array_almost_equal(
        getattr(rho_slice_pp.coords, axis),
        getattr(rho_slice_cut.coords, axis),
    )



from .analysis import (
    Coordinates,
    GasDataSet,
    GasField,
    Plotable,
)
from .from_simulation import Parameters
from .satellite import (
    NonosLick,
    compute,
    file_analysis,
    from_file,
    from_data,
    planet_analysis,
)
from .tools import find_around, find_nearest



import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import pytest
from matplotlib.colors import SymLogNorm

from nonos.api import GasDataSet
from nonos.styling import set_mpl_style


@pytest.fixture
def tmp_mpl_state():
    # reset matplotlib's state when the test is over
    style = mpl.rcParams.copy()
    yield
    mpl.rcParams.update(style)


@pytest.mark.usefixtures("tmp_mpl_state")
@pytest.mark.parametrize("scaling", [0.5, 1.0, 2.0])
@pytest.mark.mpl_image_compare(style="default")
def test_set_mpl_style(scaling):
    set_mpl_style(scaling)

    fig, ax = plt.subplots()

    x = np.linspace(0, 2 * np.pi)
    for phase in np.linspace(0, np.pi / 2, 5):
        y = np.sin(x + phase)
        ax.plot(x, y)

    ax.set(
        title="nonos style",
        xlabel="$x$ axis",
        ylabel="$y$ axis",
    )
    ax.annotate(f"{scaling=}", (0.05, 0.1), xycoords="axes fraction", fontsize=15)

    return fig


@pytest.mark.mpl_image_compare()
def test_2D_polar_plane(test_data_dir, temp_figure_and_axis):
    ds = GasDataSet(23, directory=test_data_dir / "idefix_newvtk_planet2d")
    fig, ax = temp_figure_and_axis
    ds["VX1"].map("R", "phi").plot(fig, ax, title="vr")
    return fig


@pytest.mark.mpl_image_compare()
def test_symlog(test_data_dir, temp_figure_and_axis):
    ds = GasDataSet(23, directory=test_data_dir / "idefix_newvtk_planet2d")
    fig, ax = temp_figure_and_axis
    ds["VX1"].map("R", "phi").plot(
        fig,
        ax,
        title="vr",
        norm=SymLogNorm(vmin=-0.05, vmax=0.05, linthresh=1e-4, base=10),
    )
    return fig


@pytest.mark.mpl_image_compare()
def test_3D_aa_xz(test_data_dir, temp_figure_and_axis):
    ds = GasDataSet(500, directory=test_data_dir / "idefix_spherical_planet3d")
    fig, ax = temp_figure_and_axis

    ds["RHO"].azimuthal_average().map("x", "z").plot(fig, ax, log=True, title="rho")
    return fig


@pytest.mark.mpl_image_compare()
def test_3D_vm_phiR(test_data_dir, temp_figure_and_axis):
    ds = GasDataSet(500, directory=test_data_dir / "idefix_spherical_planet3d")
    fig, ax = temp_figure_and_axis

    ds["RHO"].vertical_at_midplane().map("phi", "R").plot(
        fig, ax, log=True, title="rho"
    )
    return fig


@pytest.mark.mpl_image_compare()
def test_3D_vm_xy(test_data_dir, temp_figure_and_axis):
    ds = GasDataSet(500, directory=test_data_dir / "idefix_spherical_planet3d")
    fig, ax = temp_figure_and_axis

    ds["RHO"].vertical_at_midplane().map("x", "y").plot(fig, ax, log=True, title="rho")
    return fig


@pytest.mark.parametrize("method", ["nearest", "linear"])
@pytest.mark.mpl_image_compare()
def test_nonoslick_method(method, tmp_path, temp_figure_and_axis):
    import inifix

    from nonos.api import Coordinates, GasField, NonosLick

    fig, ax = temp_figure_and_axis

    root_size = 2
    fake_grid = {
        "geometry": "cartesian",
        "x1": np.linspace(0, 1, root_size + 1),
        "x2": np.linspace(0, 1, root_size + 1),
        "x3": np.array([1]),
    }
    fake_coords = Coordinates(**fake_grid)

    xxmed = fake_coords.xmed
    yymed = fake_coords.ymed
    xxedge = fake_coords.x
    yyedge = fake_coords.y

    rng = np.random.default_rng(seed=0)
    fake_Vx = rng.normal(0, 1, size=root_size**2).reshape(root_size, root_size)
    fake_Vy = rng.normal(0, 1, size=root_size**2).reshape(root_size, root_size)
    fake_F = rng.normal(0, 1, size=root_size**2).reshape(root_size, root_size)

    # TODO : mandatory for now to have a idefix.ini file, but should be removed in the future
    data = {
        "Output": {"vtk": 1},
        "Hydro": {},
    }
    with open(tmp_path / "idefix.ini", "wb") as fh:
        inifix.dump(data, fh)

    Vx = GasField(
        field="Vx",
        data=fake_Vx,
        coords=fake_coords,
        ngeom=fake_coords.geometry,
        on=0,
        operation="",
        directory=tmp_path,
    )
    Vy = GasField(
        field="Vy",
        data=fake_Vy,
        coords=fake_coords,
        ngeom=fake_coords.geometry,
        on=0,
        operation="",
        directory=tmp_path,
    )
    F = GasField(
        field="F",
        data=fake_F,
        coords=fake_coords,
        ngeom=fake_coords.geometry,
        on=0,
        operation="",
        directory=tmp_path,
    )
    lick = NonosLick(
        xxmed,
        yymed,
        Vx,
        Vy,
        F,
        xmin=xxedge.min(),
        xmax=xxedge.max(),
        ymin=yyedge.min(),
        ymax=yyedge.max(),
        niter_lic=1,
        size_interpolated=50 * root_size,
        method=method,
    )
    fig, ax = plt.subplots()
    lick.plot(
        fig, ax, title="F", density_streamlines=1, color_streamlines="w", cmap="inferno"
    )
    ax.set(
        title=f"{method=}",
        aspect="equal",
        xlim=(xxedge.min(), xxedge.max()),
        ylim=(yyedge.min(), yyedge.max()),
    )
    return fig



import os

import numpy as np
import numpy.testing as npt
import pytest

from nonos.api import GasDataSet, file_analysis


class TestFileAnalysis:
    @pytest.mark.parametrize(
        "directory",
        ["idefix_planet3d", "fargo_adsg_planet"],
    )
    def test_simple(self, test_data_dir, directory):
        result = file_analysis(
            "planet0.dat",
            directory=test_data_dir / directory,
        )
        assert isinstance(result, np.ndarray)

    def test_norb(self, test_data_dir):
        result = file_analysis(
            "planet0.dat",
            directory=test_data_dir / "idefix_planet3d",
            norb=10,
        )
        assert isinstance(result, np.ndarray)

    def test_norb_not_idefix(self, test_data_dir):
        with pytest.raises(NotImplementedError):
            file_analysis(
                "planet0.dat",
                directory=test_data_dir / "fargo_adsg_planet",
                norb=10,
            )

    def test_implicit_directory(self, test_data_dir):
        os.chdir(test_data_dir / "idefix_planet3d")
        result = file_analysis("planet0.dat")
        assert isinstance(result, np.ndarray)


class TestGasDataSetFromNpy:
    expected_keys = ["RHO"]
    args = (7283,)
    kwargs = {"operation": "azimuthal_average"}
    directory = "pluto_spherical"

    def test_from_npy_implicit_directory(self, test_data_dir):
        os.chdir(test_data_dir / self.directory)
        ds = GasDataSet(*self.args, **self.kwargs)
        assert sorted(ds.keys()) == self.expected_keys

    def test_from_npy_explicit_directory(self, test_data_dir):
        ds = GasDataSet(
            *self.args,
            **self.kwargs,
            directory=test_data_dir / self.directory,
        )
        assert sorted(ds.keys()) == self.expected_keys

    def test_deprecation(self, test_data_dir):
        with pytest.deprecated_call():
            ds = GasDataSet.from_npy(
                *self.args,
                **self.kwargs,
                directory=test_data_dir / self.directory,
            )
        assert sorted(ds.keys()) == self.expected_keys


def test_find_rhill(test_data_dir):
    ds = GasDataSet(23, directory=test_data_dir / "idefix_newvtk_planet2d")
    rp = ds["RHO"].find_rp()
    rhill = ds["RHO"].find_rhill()
    assert rhill < rp


def test_field_map_no_mutation(test_data_dir):
    ds = GasDataSet(500, directory=test_data_dir / "idefix_spherical_planet3d")
    f = ds["RHO"].radial_at_r(1.0).vertical_at_midplane()
    d0 = f.data.copy()
    f.map("phi", rotate_by=1.0)
    d1 = f.data.copy()
    npt.assert_array_equal(d1, d0)



import os

import matplotlib.pyplot as plt
import numpy.testing as npt
import pytest
from matplotlib.colors import SymLogNorm

from nonos.api import GasDataSet, compute, find_nearest, from_data
from nonos.main import main

ARGS_TO_CHECK = {
    "vanilla_conf": ["-geometry", "polar"],
    "diff": ["-geometry", "polar", "-diff"],
    "log": ["-geometry", "polar", "-log"],
    "movie_xy": ["-geometry", "polar", "-all", "-plane", "x", "y"],
    "movie_with_diff": ["-geometry", "polar", "-all", "-diff"],
    "movie_with_multiproc": ["-geometry", "polar", "-all", "-ncpu", "2"],
}


# CLI testing


@pytest.mark.parametrize("argv", ARGS_TO_CHECK.values(), ids=ARGS_TO_CHECK.keys())
def test_plot_simple(argv, simulation_dir, capsys, tmp_path):
    os.chdir(tmp_path)
    ret = main(argv + ["-dir", str(simulation_dir), "-geometry", "polar"])

    out, err = capsys.readouterr()
    assert err == ""
    assert out == ""
    assert ret == 0
    assert len(list(tmp_path.glob("*.png"))) > 0


@pytest.mark.parametrize("format", ["pdf", "png", "jpg"])
def test_common_image_formats(format, simulation_dir, capsys, tmp_path):
    os.chdir(tmp_path)
    ret = main(["-dir", str(simulation_dir), "-fmt", format, "-geometry", "polar"])

    out, err = capsys.readouterr()
    assert err == ""
    assert out == ""
    assert ret == 0
    assert len(list(tmp_path.glob(f"*.{format}"))) > 0


def test_plot_simple_corotation(planet_simulation_dir, capsys, tmp_path):
    os.chdir(tmp_path)
    # just check that the call returns no err
    ret = main(["-cor", "0", "-dir", str(planet_simulation_dir), "-geometry", "polar"])

    out, err = capsys.readouterr()
    assert out == ""
    assert err == ""
    assert ret == 0


def test_unknown_geometry(test_data_dir, tmp_path):
    os.chdir(tmp_path)
    with pytest.raises(
        RuntimeError, match=r"Geometry couldn't be determined from data"
    ):
        main(["-dir", str(test_data_dir / "idefix_rwi")])


def test_newvtk_geometry(test_data_dir, capsys, tmp_path):
    os.chdir(tmp_path)
    ret = main(["-cor", "0", "-dir", str(test_data_dir / "idefix_newvtk_planet2d")])
    out, err = capsys.readouterr()
    assert out == ""
    assert err == ""
    assert ret == 0


def test_verbose_info(simulation_dir, capsys, tmp_path):
    os.chdir(tmp_path)
    ret = main(["-v", "-dir", str(simulation_dir), "-geometry", "polar"])

    out, err = capsys.readouterr()
    assert err == ""
    assert "Operation took" in out
    assert "INFO" in out
    assert ret == 0


# API testing


def test_plot_planet_corotation(test_data_dir):
    os.chdir(test_data_dir / "idefix_planet3d")

    ds = GasDataSet(43, geometry="polar")
    azimfield = ds["RHO"].radial_at_r().vertical_at_midplane().map("phi").data
    assert find_nearest(azimfield, azimfield.max()) != 0

    azimfieldPlanet = (
        ds["RHO"]
        .radial_at_r()
        .vertical_at_midplane()
        .map("phi", rotate_with="planet0.dat")
        .data
    )
    assert find_nearest(azimfieldPlanet, azimfieldPlanet.max()) == 0


def test_unit_conversion(test_data_dir, temp_figure_and_axis):
    os.chdir(test_data_dir / "idefix_planet3d")

    ds = GasDataSet(43, geometry="polar")
    fig, ax = temp_figure_and_axis

    plotfield10 = (
        ds["RHO"]
        .vertical_at_midplane()
        .map("R", "phi")
        .plot(fig, ax, unit_conversion=10)
    )
    plotfield = ds["RHO"].vertical_at_midplane().map("R", "phi").plot(fig, ax)

    npt.assert_allclose(plotfield10.get_array(), 10 * plotfield.get_array())


def test_vmin_vmax_api(test_data_dir, temp_figure_and_axis):
    ds = GasDataSet(1, directory=test_data_dir / "idefix_rwi", geometry="polar")
    fig, ax = temp_figure_and_axis
    p = ds["VX1"].vertical_at_midplane().map("R", "phi")

    # check that no warning is emitted from matplotlib
    im1 = p.plot(fig, ax, vmin=-1, vmax=1, norm=SymLogNorm(linthresh=0.1, base=10))
    im2 = p.plot(fig, ax, norm=SymLogNorm(linthresh=0.1, base=10, vmin=-1, vmax=1))

    npt.assert_array_equal(im1.get_array(), im2.get_array())


def test_compute_from_data(test_data_dir):
    ne = pytest.importorskip("numexpr")

    directory = test_data_dir / "idefix_planet3d"
    os.chdir(directory)

    ds = GasDataSet(43, geometry="polar")

    rhovpfield = ds["RHO"].vertical_projection()
    vx2vpfield = ds["VX2"].vertical_projection()

    rhovp = rhovpfield.data
    vx2vp = vx2vpfield.data

    with pytest.deprecated_call():
        rhovx2_from_data = from_data(
            field="RHOVX2",
            data=rhovp * vx2vp,
            coords=rhovpfield.coords,
            on=rhovpfield.on,
            operation=rhovpfield.operation,
            directory=directory,
        )

    datane = ne.evaluate("rhovp*vx2vp")
    rhovx2_compute = compute(
        field="RHOVX2",
        data=datane,
        ref=rhovpfield,
    )

    npt.assert_array_equal(rhovx2_from_data.data, rhovx2_compute.data)


def test_pbar(simulation_dir, capsys, tmp_path):
    os.chdir(tmp_path)
    ret = main(["-pbar", "-dir", str(simulation_dir), "-geometry", "polar"])

    out, err = capsys.readouterr()
    assert err == ""
    assert out == ""
    assert ret == 0


def test_corotation_api_float(test_data_dir):
    os.chdir(test_data_dir / "idefix_newvtk_planet2d")

    ds = GasDataSet(23)
    case1 = ds["RHO"].map("x", "y", rotate_with="planet0.dat")
    ds = GasDataSet(23)
    case2 = ds["RHO"].map("x", "y", rotate_by=1.89628895460529)

    npt.assert_array_equal(case1.data, case2.data)


@pytest.mark.parametrize(
    "map_args",
    [("R", "phi"), ("phi", "R")],
)
def test_reg(test_data_dir, map_args):
    ds = GasDataSet(23, directory=test_data_dir / "idefix_newvtk_planet2d")
    fig, ax = plt.subplots()
    ds["RHO"].map(*map_args).plot(fig, ax)



import os
import sys

import pytest

from nonos.api import Parameters


def test_init_params_wo_a_file(tmp_path):
    os.chdir(tmp_path)
    with (
        pytest.deprecated_call(),
        pytest.raises(
            FileNotFoundError, match=r"idefix.ini, pluto.ini, variables.par not found"
        ),
    ):
        Parameters()


@pytest.mark.skipif(sys.platform.startswith("win"), reason="does not run on windows")
def test_config_dir_not_found(tmp_path):
    with pytest.deprecated_call(), pytest.raises(FileNotFoundError, match=r"not found"):
        Parameters(inifile="notafile", code="idefix", directory=tmp_path)


def test_config_inifile_but_nocode(tmp_path):
    with (
        pytest.deprecated_call(),
        pytest.raises(ValueError, match=r"both inifile and code have to be given"),
    ):
        Parameters(inifile="notafile", directory=tmp_path)



import re
from pathlib import Path

from nonos.config import DEFAULTS

SOURCE_DIR = Path(__file__).parents[1]
with open(SOURCE_DIR.joinpath("nonos", "main.py")) as fh:
    source = fh.read()

used_keys = set(re.findall(r"""args\[['"](\w+)['"]\]""", source))
def_keys = set(DEFAULTS)


# check that every key used is defined
def test_no_undef_key():
    if missing := used_keys.difference(def_keys):  # pragma: no cover
        msg = f"The following {len(missing)} keys are used but not defined:\n"
        msg += "\n".join(list(missing))
    assert not missing, msg


def test_no_unused_key():
    if missing := def_keys.difference(used_keys):  # pragma: no cover
        msg = f"The following {len(missing)} keys are defined but not used:\n"
        msg += "\n".join(list(missing))
    assert not missing, msg



__all__ = [
    "ReaderMixin",
]


class ReaderMixin:
    __slots__: list[str] = []

    def __init__(self, *args, **kwargs):  # noqa: ARG002
        raise TypeError(f"{self.__class__} is not instantiable")



#!/usr/bin/env python
"""
Analysis tool for idefix/pluto/fargo3d simulations (in polar coordinates).
"""
# adapted from pbenitez-llambay, gwafflard-fernandez, cmt robert & glesur

import argparse
import functools
import importlib.resources as importlib_resources
import os
import re
import time
from collections import ChainMap
from importlib.metadata import version
from multiprocessing import Pool
from pathlib import Path
from typing import Any, Optional

import cblind  # noqa
import inifix
import numpy as np
from packaging.version import Version

from nonos.__version__ import __version__
from nonos.api import GasDataSet
from nonos.api._angle_parsing import _parse_planet_file
from nonos.config import DEFAULTS
from nonos.loaders import loader_from
from nonos.logging import (
    configure_logger,
    logger,
    parse_verbose_level,
    print_err,
    print_warn,
)
from nonos.parsing import (
    is_set,
    parse_image_format,
    parse_output_number_range,
    parse_range,
    range_converter,
    userval_or_default,
)
from nonos.styling import set_mpl_style

INIFIX_GE_5_0 = Version(version("inifix")) >= Version("5.0.0")


# process function for parallelisation purpose with progress bar
# counterParallel = Value('i', 0) # initialization of a counter
def process_field(
    on,
    operations: list[str],
    field,
    plane,
    geometry,
    diff,
    log,
    planet_file: Optional[str],
    extent,
    vmin,
    vmax,
    scaling: float,
    cmap,
    title,
    unit_conversion: int,
    datadir,
    show: bool,
    dpi: int,
    fmt: str,
    theta,
    z,
    phi,
    distance,
    *,
    log_level,
):
    import matplotlib.pyplot as plt

    configure_logger(level=log_level)
    set_mpl_style(scaling=scaling)
    if geometry == "unset":
        geometry = None
    ds = GasDataSet(on, geometry=geometry, directory=datadir)
    dsop = ds[field]
    if diff:
        dsop = dsop.diff(0)
    if "vm" in operations:
        dsop = dsop.vertical_at_midplane()
    elif "vp" in operations:
        dsop = dsop.vertical_projection(z=z)
    elif "lt" in operations:
        dsop = dsop.latitudinal_at_theta(theta=theta)
    elif "lp" in operations:
        dsop = dsop.latitudinal_projection(theta=theta)
    elif "vz" in operations:
        dsop = dsop.vertical_at_z(z=z)

    if "ap" in operations:
        dsop = dsop.azimuthal_at_phi(phi=phi)
    elif "apl" in operations:
        dsop = dsop.azimuthal_at_planet(planet_file=planet_file)
    elif "aa" in operations:
        dsop = dsop.azimuthal_average()

    if "rr" in operations:
        dsop = dsop.radial_at_r(distance=distance)

    logger.debug("operations performed: {}", operations)

    dim = 3 - dsop.shape.count(1)
    logger.debug("plotting a {}D plot.", dim)

    if plane is None:
        dsop_dict = dsop.coords.get_attributes
        default_plane = []
        for key, val in dsop_dict.items():
            if not isinstance(val, str) and val.shape[0] > 2:
                default_plane.append(key)
        # default_plane = ["x","y"]
        plane = default_plane

    fig = plt.figure()
    ax = fig.add_subplot(111, polar=False)
    if dim == 1:
        dsop.map(plane[0], rotate_with=planet_file).plot(
            fig,
            ax,
            log=log,
            vmin=vmin,
            vmax=vmax,
            title=f"${title}$",
            unit_conversion=unit_conversion,
        )
        akey = dsop.map(plane[0], rotate_with=planet_file).dict_plotable["abscissa"]
        avalue = dsop.map(plane[0], rotate_with=planet_file).dict_plotable[akey]
        extent = parse_range(extent, dim=dim)
        extent = range_converter(extent, abscissa=avalue, ordinate=np.zeros(2))
        ax.set_xlim(extent[0], extent[1])
    elif dim == 2:
        dsop.map(plane[0], plane[1], rotate_with=planet_file).plot(
            fig,
            ax,
            log=log,
            vmin=vmin,
            vmax=vmax,
            cmap=cmap,
            title=f"${title}$",
            unit_conversion=unit_conversion,
        )
        akey = dsop.map(plane[0], plane[1], rotate_with=planet_file).dict_plotable[
            "abscissa"
        ]
        okey = dsop.map(plane[0], plane[1], rotate_with=planet_file).dict_plotable[
            "ordinate"
        ]
        avalue = dsop.map(plane[0], plane[1], rotate_with=planet_file).dict_plotable[
            akey
        ]
        ovalue = dsop.map(plane[0], plane[1], rotate_with=planet_file).dict_plotable[
            okey
        ]
        extent = parse_range(extent, dim=dim)
        extent = range_converter(
            extent,
            abscissa=avalue,
            ordinate=ovalue,
        )
        ax.set_xlim(extent[0], extent[1])
        ax.set_ylim(extent[2], extent[3])

    logger.debug("processed the data before plotting.")

    if "x" and "y" in plane:
        ax.set_aspect("equal")

    if show:
        plt.show()
    else:
        logger.debug("saving plot: started")
        filename = f"{''.join(plane)}_{field}_{'_'.join(operations)}{'_diff' if diff else '_'}{'_log' if log else ''}{on:04d}.{fmt}"
        fig.savefig(filename, bbox_inches="tight", dpi=dpi)
        logger.debug("saving plot: finished ({})", filename)

    plt.close(fig)


def get_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        prog="nonos",
        description=__doc__,
    )

    parser.add_argument(
        "-dir",
        dest="datadir",
        help=f"location of output files and param files (default: '{DEFAULTS['datadir']}').",
    )
    parser.add_argument(
        "-field",
        # choices=["RHO", "VX1", "VX2", "VX3", "BX1", "BX2", "BX3", "PRS"],
        help=f"name of field to plot (default: '{DEFAULTS['field']}').",
    )
    parser.add_argument(
        "-geometry",
        type=str,
        choices=["polar", "cylindrical", "spherical", "cartesian"],
        help=f"if the geometry of idefix outputs is not recognized  (default: '{DEFAULTS['geometry']}').",
    )
    parser.add_argument(
        "-operation",
        type=str,
        nargs="+",
        choices=["vm", "vp", "vz", "lt", "lp", "aa", "ap", "apl", "rr"],
        help=f"operation to apply to the fild (default: '{DEFAULTS['operation']}').",
    )
    parser.add_argument(
        "-plane",
        type=str,
        nargs="+",
        help=f"abscissa and ordinate of the plane of projection (default: '{DEFAULTS['plane']}'), example: r phi",
    )

    # TODO(GWF): add support for -rotate_by -1.2453036989845032 (idefix_newvtk_planet2d)
    parser.add_argument(
        "-corotate",
        type=int,
        default=None,
        help="planet number that defines with which planet the grid corotates.",
    )
    parser.add_argument(
        "-range",
        type=str,
        nargs="+",
        help=f"range of matplotlib window (default: {DEFAULTS['range']}), example: x x -2 2",
    )
    parser.add_argument(
        "-vmin",
        type=float,
        help=f"min value (default: {DEFAULTS['vmin']})",
    )
    parser.add_argument(
        "-vmax",
        type=float,
        help=f"max value (default: {DEFAULTS['vmax']})",
    )
    parser.add_argument(
        "-theta",
        type=float,
        help=f"if latitudinal operation (default: {DEFAULTS['theta']})",
    )
    parser.add_argument(
        "-z",
        type=float,
        help=f"if vertical operation (default: {DEFAULTS['z']})",
    )
    parser.add_argument(
        "-phi",
        type=float,
        help=f"if azimuthal operation (default: {DEFAULTS['phi']})",
    )
    parser.add_argument(
        "-distance",
        type=float,
        help=f"if radial operation (default: {DEFAULTS['distance']})",
    )
    parser.add_argument(
        "-cpu",
        "-ncpu",
        dest="ncpu",
        type=int,
        help=f"number of parallel processes (default: {DEFAULTS['ncpu']}).",
    )

    select_group = parser.add_mutually_exclusive_group()
    select_group.add_argument(
        "-on",
        type=int,
        nargs="+",
        help="output number(s) (on) to plot. "
        "This can be a single value or a range (start, end, [step]) where both ends are inclusive. "
        "(default: last output available).",
    )
    select_group.add_argument(
        "-all",
        action="store_true",
        help="save an image for every available snapshot (this will force show=False).",
    )

    # boolean flags use False as a default value (...by default)
    # forcing them to None instead will allow them to be pruned
    # when we build the ChainMap config
    flag_group = parser.add_argument_group("boolean flags")
    flag_group.add_argument(
        "-diff",
        action="store_true",
        default=None,
        help="plot the relative perturbation of the field f, i.e. (f-f0)/f0.",
    )
    flag_group.add_argument(
        "-log",
        action="store_true",
        default=None,
        help="plot the log10 of the field f, i.e. log(f).",
    )
    flag_group.add_argument(
        "-pbar",
        dest="progressBar",
        action="store_true",
        default=None,
        help="display a progress bar",
    )

    parser.add_argument(
        "-scaling",
        dest="scaling",
        type=float,
        help=f"scale the overall sizes of features in the graph (fonts, linewidth...) (default: {DEFAULTS['scaling']}).",
    )
    parser.add_argument(
        "-cmap",
        help=f"choice of colormap for the 2D maps (default: '{DEFAULTS['cmap']}').",
    )
    parser.add_argument(
        "-title",
        type=str,
        help=f"name of the field in the colorbar for the 2D maps (default: '{DEFAULTS['title']}').",
    )
    parser.add_argument(
        "-uc",
        "-unit_conversion",
        dest="unit_conversion",
        type=float,
        help=f"conversion factor for the considered quantity (default: '{DEFAULTS['unit_conversion']}').",
    )
    parser.add_argument(
        "-fmt",
        "-format",
        dest="format",
        help=f"select output image file format (default: {DEFAULTS['format']})",
    )
    parser.add_argument(
        "-dpi",
        type=int,
        help=f"image file resolution (default: {DEFAULTS['dpi']})",
    )

    cli_only_group = parser.add_argument_group("CLI-only options")
    cli_input_group = cli_only_group.add_mutually_exclusive_group()
    cli_input_group.add_argument(
        "-input", "-i", dest="input", type=str, help="specify a configuration file."
    )
    cli_input_group.add_argument(
        "-isolated", action="store_true", help="ignore any existing 'nonos.ini' file."
    )
    cli_action_group = cli_only_group.add_mutually_exclusive_group()
    cli_action_group.add_argument(
        "-d",
        "-display",
        dest="display",
        action="store_true",
        help="open a graphic window with the plot (only works with a single image)",
    )
    cli_action_group.add_argument(
        "-version",
        "--version",
        action="store_true",
        help="show raw version number and exit",
    )
    cli_action_group.add_argument(
        "-logo",
        action="store_true",
        help="show Nonos logo with version number, and exit.",
    )
    cli_action_group.add_argument(
        "-config", action="store_true", help="show configuration and exit."
    )

    cli_debug_group = cli_only_group.add_mutually_exclusive_group()
    cli_debug_group.add_argument(
        "-v",
        "-verbose",
        "--verbose",
        action="count",
        default=0,
        help="increase output verbosity (-v: info, -vv: debug).",
    )
    return parser


def main(argv: Optional[list[str]] = None) -> int:
    parser = get_parser()
    clargs = vars(parser.parse_args(argv))

    # special cases: destructively consume CLI-only arguments with dict.pop

    if clargs.pop("logo"):
        logo = importlib_resources.files("nonos").joinpath("logo.txt").read_text()
        print(f"{logo}{__doc__}Version {__version__}")
        return 0

    if clargs.pop("version"):
        print(__version__)
        return 0

    # clargs.pop("verbose")
    level = parse_verbose_level(clargs.pop("verbose"))
    configure_logger(level=level)
    # logger.setLevel(level)

    if clargs.pop("isolated"):
        config_file_args: dict[str, Any] = {}
    elif (ifile := clargs.pop("input")) is not None:
        if not Path(ifile).is_file():
            print_err(f"Couldn't find requested input file {ifile!r}.")
            return 1
        print_warn(f"[bold white]Using parameters from {ifile!r}.")
        config_file_args = inifix.load(ifile)
    elif Path("nonos.ini").is_file():
        print_warn("[bold white]Using parameters from 'nonos.ini'.")
        config_file_args = inifix.load("nonos.ini")
    else:
        config_file_args = {}

    # check that every parameter in the configuration is also exposed to the CLI
    assert not set(DEFAULTS).difference(set(clargs))

    # squeeze out any unset value form cli config to leave room for file parameters
    clargs = {k: v for k, v in clargs.items() if v is not None}

    # NOTE: init.config is also a ChainMap instance with a default layer
    # this may be seen either as hyperstatism (good thing) or error prone redundancy (bad thing)
    args = ChainMap(clargs, config_file_args, DEFAULTS)
    if clargs.pop("config"):
        conf_repr = {}
        for key in DEFAULTS:
            conf_repr[key] = args[key]
        print(f"# Generated with nonos {__version__}")
        s = inifix.dumps(conf_repr)
        if INIFIX_GE_5_0:
            print(inifix.format_string(s))  # type: ignore [attr-defined]
        else:
            from inifix.format import iniformat

            print(iniformat(s))
        return 0

    try:
        loader = loader_from(directory=args["datadir"])
    except (FileNotFoundError, RuntimeError, ValueError) as exc:
        print_err(exc)
        return 1

    data_files = loader.binary_reader.get_bin_files(args["datadir"])

    available = set()
    for fn in data_files:
        if (num := re.search(r"\d+", fn.name)) is not None:
            available.add(int(num.group()))

    if args.pop("all"):
        requested = available
    else:
        try:
            requested = set(
                parse_output_number_range(args["on"], maxval=max(available))
            )
        except ValueError as exc:
            print_err(exc)
            return 1

    if not (toplot := list(requested.intersection(available))):
        print_err(
            f"No requested output file was found (requested {requested}, found {available})."
        )
        return 1
    args["on"] = toplot

    if (show := clargs.pop("display")) and len(args["on"]) > 1:
        print_warn("display mode can not be used with multiple images, turning it off.")
        show = False

    if not show:
        try:
            args["format"] = parse_image_format(args["format"])
        except ValueError as exc:
            print_err(exc)
            return 1

    # check that every CLI-only argument was consumed at this point
    assert not set(clargs).difference(set(DEFAULTS))

    args["field"] = args["field"].upper()
    extent = args["range"]

    if args["ncpu"] > (ncpu := min(args["ncpu"], os.cpu_count())):
        print_warn(
            f"Requested {args['ncpu']}, but the runner only has access to {ncpu}."
        )

    if args["progressBar"]:
        from rich.progress import track

        def mytrack(iterable, *args, **kwargs):
            return track(iterable, *args, **kwargs)

    else:
        # replace rich.progress.track with a no-op dummy
        def mytrack(iterable, *args, **kwargs):  # noqa: ARG001
            return iterable

    planet_file: Optional[str]
    if not is_set(args["corotate"]):
        planet_file = None
    else:
        planet_file = _parse_planet_file(planet_number=args["corotate"])

    # call of the process_field function, whether it be in parallel or not
    # TODO: reduce this to the bare minimum
    func_kwargs = dict(  # noqa: C408
        operations=userval_or_default(args["operation"], default=["vm"]),
        field=args["field"],
        plane=userval_or_default(args["plane"], default=None),
        geometry=userval_or_default(args["geometry"], default="unset"),
        diff=args["diff"],
        log=args["log"],
        planet_file=planet_file,
        extent=extent,
        vmin=userval_or_default(args["vmin"], default=None),
        vmax=userval_or_default(args["vmax"], default=None),
        scaling=args["scaling"],
        cmap=args["cmap"],
        title=userval_or_default(args["title"], default=args["field"]),
        unit_conversion=args["unit_conversion"],
        datadir=args["datadir"],
        show=show,
        dpi=args["dpi"],
        fmt=args["format"],
        theta=userval_or_default(args["theta"], default=None),
        z=userval_or_default(args["z"], default=None),
        phi=userval_or_default(args["phi"], default=None),
        distance=userval_or_default(args["distance"], default=None),
        log_level=level,
    )

    logger.info("Starting main loop")
    tstart = time.time()
    if ncpu == 1:
        for on in args["on"]:
            process_field(on, **func_kwargs)
    else:
        func = functools.partial(process_field, **func_kwargs)
        with Pool(ncpu) as pool:
            list(
                mytrack(
                    pool.imap(func, args["on"]),
                    description="Processing snapshots",
                    total=len(args["on"]),
                )
            )
    if not show:
        logger.info("Operation took {:.2f}s", time.time() - tstart)

    return 0



import sys

if sys.version_info >= (3, 11):
    pass
else:
    from enum import Enum

    # vendored from Python 3.11.0
    class ReprEnum(Enum):
        """
        Only changes the repr(), leaving str() and format() to the mixed-in type.
        """

    class StrEnum(str, ReprEnum):
        """
        Enum where members are also (and must be) strings
        """

        def __new__(cls, *values):
            "values must already be of type `str`"
            if len(values) > 3:
                raise TypeError(f"too many arguments for str(): {values!r}")
            if len(values) == 1:
                # it must be a string
                if not isinstance(values[0], str):
                    raise TypeError(f"{values[0]!r} is not a string")
            if len(values) >= 2:
                # check that encoding argument is a string
                if not isinstance(values[1], str):
                    raise TypeError(f"encoding must be a string, not {values[1]!r}")
            if len(values) == 3:
                # check that errors argument is a string
                if not isinstance(values[2], str):
                    raise TypeError(f"errors must be a string, not {values[2]!r}")
            value = str(*values)
            member = str.__new__(cls, value)
            member._value_ = value
            return member

        def _generate_next_value_(name, start, count, last_values):  # noqa: ARG002
            """
            Return the lower-cased version of the member name.
            """
            return name.lower()



from typing import Any, Literal, Optional, TypeVar, Union, overload

import numpy as np


def is_set(x: Any) -> bool:
    return x not in (None, "unset")


T1 = TypeVar("T1")
T2 = TypeVar("T2")


def userval_or_default(userval: T1, /, *, default: T2) -> Union[T1, T2]:
    # it'd be nice to avoid a Union as a return type, however it's not clear
    # how to express what this function does in typing language.
    # In practice this is used in places where it's very hard to constrain T1, so
    # the infered return type always degrades down to Any anyway.
    if is_set(userval):
        return userval
    else:
        return default


def parse_output_number_range(
    on: Union[list[int], int, Literal["unset"], None],
    maxval: Optional[int] = None,
) -> list[int]:
    if not is_set(on):
        if maxval is None:
            raise ValueError("Can't parse a range from unset values without a max.")
        return [maxval]

    if isinstance(on, int):
        return [on]

    assert isinstance(on, list) and all(isinstance(o, int) for o in on)

    if len(on) > 3:
        raise ValueError(
            f"Can't parse a range from sequence {on} with more than 3 values."
        )
    if len(on) == 1:
        return on

    if on[1] < on[0]:
        raise ValueError("Can't parse a range with max < min.")

    # make the upper boundary inclusive
    on[1] += 1
    ret = list(range(*on))
    if maxval is not None and (max_requested := ret[-1]) > maxval:
        raise ValueError(
            f"No output beyond {maxval} is available, but {max_requested} was requested."
        )
    return ret


def parse_range(extent, dim: int) -> tuple[Optional[float], ...]:
    if not is_set(extent):
        return (None,) * 2 * dim

    if len(extent) != 2 * dim:
        raise ValueError(
            f"Received sequence `extent` with incorrect size {len(extent)}. Expected exactly {2*dim=} values."
        )
    return tuple(float(i) if i != "x" else None for i in extent)


@overload
def range_converter(
    extent: tuple[Optional[float], Optional[float]],
    abscissa: np.ndarray,
    ordinate: np.ndarray,
) -> tuple[float, float]: ...


@overload
def range_converter(
    extent: tuple[Optional[float], Optional[float], Optional[float], Optional[float]],
    abscissa: np.ndarray,
    ordinate: np.ndarray,
) -> tuple[float, float, float, float]: ...


def range_converter(extent, abscissa, ordinate):
    if len(extent) == 4:
        return (
            _ if (_ := extent[0]) is not None else abscissa.min(),
            _ if (_ := extent[1]) is not None else abscissa.max(),
            _ if (_ := extent[2]) is not None else ordinate.min(),
            _ if (_ := extent[3]) is not None else ordinate.max(),
        )
    elif len(extent) == 2:
        return (
            _ if (_ := extent[0]) is not None else abscissa.min(),
            _ if (_ := extent[1]) is not None else abscissa.max(),
        )
    else:
        raise TypeError(f"Expected extent to be of lenght 2 or 4, got {len(extent)=}")


def parse_image_format(s: Optional[str]) -> str:
    from matplotlib.backend_bases import FigureCanvasBase

    if not is_set(s):
        return FigureCanvasBase.get_default_filetype()

    assert isinstance(s, str)
    _, _, ext = s.rpartition(".")
    if ext not in (
        available := list(FigureCanvasBase.get_supported_filetypes().keys())
    ):
        raise ValueError(
            f"Received unknown file format '{s}'. "
            f"Available formated are {available}."
        )
    return ext
